from app.filter import clean_query
from app.request import send_tor_signal
from app.utils.session import generate_key
from app.utils.bangs import gen_bangs_json, load_all_bangs
from app.utils.misc import gen_file_hash, read_config_bool
from app.utils.ua_generator import load_ua_pool
from base64 import b64encode
from bs4 import MarkupResemblesLocatorWarning
from datetime import datetime, timedelta
from dotenv import load_dotenv
from flask import Flask
import json
import logging.config
import os
import sys
from stem import Signal
import threading
import warnings

from werkzeug.middleware.proxy_fix import ProxyFix

from app.services.http_client import HttpxClient
from app.services.provider import close_all_clients
from app.version import __version__

app = Flask(__name__, static_folder=os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'static'))

app.wsgi_app = ProxyFix(app.wsgi_app)

# look for WHOOGLE_ENV, else look in parent directory
dot_env_path = os.getenv(
    "WHOOGLE_DOTENV_PATH",
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "../whoogle.env"))

# Load .env file if enabled
if os.path.exists(dot_env_path):
    load_dotenv(dot_env_path)

app.enc_key = generate_key()

if read_config_bool('HTTPS_ONLY'):
    app.config['SESSION_COOKIE_NAME'] = '__Secure-session'
    app.config['SESSION_COOKIE_SECURE'] = True

app.config['VERSION_NUMBER'] = __version__
app.config['APP_ROOT'] = os.getenv(
    'APP_ROOT',
    os.path.dirname(os.path.abspath(__file__)))
app.config['STATIC_FOLDER'] = os.getenv(
    'STATIC_FOLDER',
    os.path.join(app.config['APP_ROOT'], 'static'))
app.config['BUILD_FOLDER'] = os.path.join(
    app.config['STATIC_FOLDER'], 'build')
app.config['CACHE_BUSTING_MAP'] = {}
app.config['BUNDLE_STATIC'] = read_config_bool('WHOOGLE_BUNDLE_STATIC')
with open(os.path.join(app.config['STATIC_FOLDER'], 'settings/languages.json'), 'r', encoding='utf-8') as f:
    app.config['LANGUAGES'] = json.load(f)
with open(os.path.join(app.config['STATIC_FOLDER'], 'settings/countries.json'), 'r', encoding='utf-8') as f:
    app.config['COUNTRIES'] = json.load(f)
with open(os.path.join(app.config['STATIC_FOLDER'], 'settings/time_periods.json'), 'r', encoding='utf-8') as f:
    app.config['TIME_PERIODS'] = json.load(f)
with open(os.path.join(app.config['STATIC_FOLDER'], 'settings/translations.json'), 'r', encoding='utf-8') as f:
    app.config['TRANSLATIONS'] = json.load(f)
with open(os.path.join(app.config['STATIC_FOLDER'], 'settings/themes.json'), 'r', encoding='utf-8') as f:
    app.config['THEMES'] = json.load(f)
with open(os.path.join(app.config['STATIC_FOLDER'], 'settings/header_tabs.json'), 'r', encoding='utf-8') as f:
    app.config['HEADER_TABS'] = json.load(f)
app.config['CONFIG_PATH'] = os.getenv(
    'CONFIG_VOLUME',
    os.path.join(app.config['STATIC_FOLDER'], 'config'))
app.config['DEFAULT_CONFIG'] = os.path.join(
    app.config['CONFIG_PATH'],
    'config.json')
app.config['CONFIG_DISABLE'] = read_config_bool('WHOOGLE_CONFIG_DISABLE')
app.config['SESSION_FILE_DIR'] = os.path.join(
    app.config['CONFIG_PATH'],
    'session')
# Maximum session file size in bytes (4KB limit to prevent abuse and disk exhaustion)
# Session files larger than this are ignored during cleanup to avoid processing
# potentially malicious or corrupted files
app.config['MAX_SESSION_SIZE'] = 4000
app.config['BANG_PATH'] = os.getenv(
    'CONFIG_VOLUME',
    os.path.join(app.config['STATIC_FOLDER'], 'bangs'))
app.config['BANG_FILE'] = os.path.join(
    app.config['BANG_PATH'],
    'bangs.json')

# Global services registry (simple DI)
app.services = {}


@app.teardown_appcontext
def _teardown_clients(exception):
    try:
        close_all_clients()
    except Exception:
        pass

# Ensure all necessary directories exist
if not os.path.exists(app.config['CONFIG_PATH']):
    os.makedirs(app.config['CONFIG_PATH'])

if not os.path.exists(app.config['SESSION_FILE_DIR']):
    os.makedirs(app.config['SESSION_FILE_DIR'])

if not os.path.exists(app.config['BANG_PATH']):
    os.makedirs(app.config['BANG_PATH'])

if not os.path.exists(app.config['BUILD_FOLDER']):
    os.makedirs(app.config['BUILD_FOLDER'])

# Initialize User Agent pool
app.config['UA_CACHE_PATH'] = os.path.join(app.config['CONFIG_PATH'], 'ua_cache.json')
try:
    app.config['UA_POOL'] = load_ua_pool(app.config['UA_CACHE_PATH'], count=10)
except Exception as e:
    # If UA pool loading fails, log warning and set empty pool
    # The gen_user_agent function will handle the fallback
    print(f"Warning: Could not initialize UA pool: {e}")
    app.config['UA_POOL'] = []

# Session values - Secret key management
# Priority: environment variable → file → generate new
def get_secret_key():
    """Load or generate secret key with validation.
    
    Priority order:
    1. WHOOGLE_SECRET_KEY environment variable
    2. Existing key file
    3. Generate new key and save to file
    
    Returns:
        str: Valid secret key for Flask sessions
    """
    # Check environment variable first
    env_key = os.getenv('WHOOGLE_SECRET_KEY', '').strip()
    if env_key:
        # Validate env key has minimum length
        if len(env_key) >= 32:
            return env_key
        else:
            print(f"Warning: WHOOGLE_SECRET_KEY too short ({len(env_key)} chars, need 32+). Using file/generated key instead.", file=sys.stderr)
    
    # Check file-based key
    app_key_path = os.path.join(app.config['CONFIG_PATH'], 'whoogle.key')
    if os.path.exists(app_key_path):
        try:
            with open(app_key_path, 'r', encoding='utf-8') as f:
                key = f.read().strip()
                # Validate file key
                if len(key) >= 32:
                    return key
                else:
                    print(f"Warning: Key file too short, regenerating", file=sys.stderr)
        except (PermissionError, IOError) as e:
            print(f"Warning: Could not read key file: {e}", file=sys.stderr)
    
    # Generate new key
    new_key = str(b64encode(os.urandom(32)))
    try:
        with open(app_key_path, 'w', encoding='utf-8') as key_file:
            key_file.write(new_key)
    except (PermissionError, IOError) as e:
        print(f"Warning: Could not save key file: {e}. Key will not persist across restarts.", file=sys.stderr)
    
    return new_key

app.config['SECRET_KEY'] = get_secret_key()
app.config['PERMANENT_SESSION_LIFETIME'] = timedelta(days=365)

# NOTE: SESSION_COOKIE_SAMESITE must be set to 'lax' to allow the user's
# previous session to persist when accessing the instance from an external
# link. Setting this value to 'strict' causes Whoogle to revalidate a new
# session, and fail, resulting in cookies being disabled.
app.config['SESSION_COOKIE_SAMESITE'] = 'Lax'

# Config fields that are used to check for updates
app.config['RELEASES_URL'] = 'https://github.com/' \
                             'benbusby/whoogle-search/releases'
app.config['LAST_UPDATE_CHECK'] = datetime.now() - timedelta(hours=24)
app.config['HAS_UPDATE'] = ''

# The alternative to Google Translate is treated a bit differently than other
# social media site alternatives, in that it is used for any translation
# related searches.
translate_url = os.getenv('WHOOGLE_ALT_TL', 'https://farside.link/lingva')
if not translate_url.startswith('http'):
    translate_url = 'https://' + translate_url
app.config['TRANSLATE_URL'] = translate_url

app.config['CSP'] = 'default-src \'none\';' \
                    'frame-src ' + translate_url + ';' \
                    'manifest-src \'self\';' \
                    'img-src \'self\' data:;' \
                    'style-src \'self\' \'unsafe-inline\';' \
                    'script-src \'self\';' \
                    'media-src \'self\';' \
                    'connect-src \'self\';'

# Generate DDG bang filter
generating_bangs = False
if not os.path.exists(app.config['BANG_FILE']):
    generating_bangs = True
    with open(app.config['BANG_FILE'], 'w', encoding='utf-8') as f:
        json.dump({}, f)
    bangs_thread = threading.Thread(
        target=gen_bangs_json,
        args=(app.config['BANG_FILE'],))
    bangs_thread.start()

# Build new mapping of static files for cache busting
cache_busting_dirs = ['css', 'js']
for cb_dir in cache_busting_dirs:
    full_cb_dir = os.path.join(app.config['STATIC_FOLDER'], cb_dir)
    for cb_file in os.listdir(full_cb_dir):
        # Create hash from current file state
        full_cb_path = os.path.join(full_cb_dir, cb_file)
        cb_file_link = gen_file_hash(full_cb_dir, cb_file)
        build_path = os.path.join(app.config['BUILD_FOLDER'], cb_file_link)

        try:
            os.symlink(full_cb_path, build_path)
        except FileExistsError:
            # Symlink hasn't changed, ignore
            pass

        # Create mapping for relative path urls
        map_path = build_path.replace(app.config['APP_ROOT'], '')
        if map_path.startswith('/'):
            map_path = map_path[1:]
        app.config['CACHE_BUSTING_MAP'][cb_file] = map_path

# Optionally create simple bundled assets (opt-in via WHOOGLE_BUNDLE_STATIC=1)
if app.config['BUNDLE_STATIC']:
    # CSS bundle: include all css except theme files (end with -theme.css)
    css_dir = os.path.join(app.config['STATIC_FOLDER'], 'css')
    css_parts = []
    for name in sorted(os.listdir(css_dir)):
        if not name.endswith('.css'):
            continue
        if name.endswith('-theme.css'):
            continue
        try:
            with open(os.path.join(css_dir, name), 'r', encoding='utf-8') as f:
                css_parts.append(f.read())
        except Exception:
            pass
    css_bundle = '\n'.join(css_parts)
    if css_bundle:
        css_tmp = os.path.join(app.config['BUILD_FOLDER'], 'app.css')
        with open(css_tmp, 'w', encoding='utf-8') as f:
            f.write(css_bundle)
        css_hashed = gen_file_hash(app.config['BUILD_FOLDER'], 'app.css')
        os.replace(css_tmp, os.path.join(app.config['BUILD_FOLDER'], css_hashed))
        map_path = os.path.join('app/static/build', css_hashed)
        app.config['CACHE_BUSTING_MAP']['bundle.css'] = map_path

    # JS bundle: include all js files
    js_dir = os.path.join(app.config['STATIC_FOLDER'], 'js')
    js_parts = []
    for name in sorted(os.listdir(js_dir)):
        if not name.endswith('.js'):
            continue
        try:
            with open(os.path.join(js_dir, name), 'r', encoding='utf-8') as f:
                js_parts.append(f.read())
        except Exception:
            pass
    js_bundle = '\n;'.join(js_parts)
    if js_bundle:
        js_tmp = os.path.join(app.config['BUILD_FOLDER'], 'app.js')
        with open(js_tmp, 'w', encoding='utf-8') as f:
            f.write(js_bundle)
        js_hashed = gen_file_hash(app.config['BUILD_FOLDER'], 'app.js')
        os.replace(js_tmp, os.path.join(app.config['BUILD_FOLDER'], js_hashed))
        map_path = os.path.join('app/static/build', js_hashed)
        app.config['CACHE_BUSTING_MAP']['bundle.js'] = map_path

# Templating functions
app.jinja_env.globals.update(clean_query=clean_query)
app.jinja_env.globals.update(
    cb_url=lambda f: app.config['CACHE_BUSTING_MAP'][f.lower()])
app.jinja_env.globals.update(
    bundle_static=lambda: app.config.get('BUNDLE_STATIC', False))

# Attempt to acquire tor identity, to determine if Tor config is available
send_tor_signal(Signal.HEARTBEAT)

# Suppress spurious warnings from BeautifulSoup
warnings.simplefilter('ignore', MarkupResemblesLocatorWarning)

from app import routes  # noqa

# The gen_bangs_json function takes care of loading bangs, so skip it here if
# it's already being loaded
if not generating_bangs:
    load_all_bangs(app.config['BANG_FILE'])

# Disable logging from imported modules
logging.config.dictConfig({
    'version': 1,
    'disable_existing_loggers': True,
})



import json
import types

import pytest

from app.models.endpoint import Endpoint
from app.utils import search as search_mod


@pytest.fixture
def stubbed_search_response(monkeypatch):
    # Stub Search.new_search_query to return a stable query
    def fake_new_query(self):
        self.query = 'whoogle'
        return self.query

    # Return a minimal filtered HTML snippet with a couple of links
    html = (
        '<div id="main">'
        '  <a href="https://example.com/page">Example Page</a>'
        '  <a href="/relative">Relative</a>'
        '  <a href="https://example.org/other">Other</a>'
        '</div>'
    )

    def fake_generate(self):
        return html

    monkeypatch.setattr(search_mod.Search, 'new_search_query', fake_new_query)
    monkeypatch.setattr(search_mod.Search, 'generate_response', fake_generate)


def test_search_json_accept(client, stubbed_search_response):
    rv = client.get(f'/{Endpoint.search}?q=whoogle', headers={'Accept': 'application/json'})
    assert rv._status_code == 200
    data = json.loads(rv.data)
    assert data['query'] == 'whoogle'
    assert isinstance(data['results'], list)
    hrefs = {item['href'] for item in data['results']}
    assert 'https://example.com/page' in hrefs
    assert 'https://example.org/other' in hrefs
    # Relative href should be excluded
    assert not any(href.endswith('/relative') for href in hrefs)
    # Verify new fields are present while maintaining backward compatibility
    for result in data['results']:
        assert 'href' in result
        assert 'text' in result  # Original field maintained
        assert 'title' in result  # New field
        assert 'content' in result  # New field


def test_search_json_format_param(client, stubbed_search_response):
    rv = client.get(f'/{Endpoint.search}?q=whoogle&format=json')
    assert rv._status_code == 200
    data = json.loads(rv.data)
    assert data['query'] == 'whoogle'
    assert len(data['results']) >= 2


def test_search_json_feeling_lucky(client, monkeypatch):
    # Force query to be interpreted as feeling lucky and return a redirect URL
    def fake_new_query(self):
        self.query = 'whoogle !'
        # emulate behavior of new_search_query setting feeling_lucky
        self.feeling_lucky = True
        return self.query

    def fake_generate(self):
        return 'https://example.com/lucky'

    monkeypatch.setattr(search_mod.Search, 'new_search_query', fake_new_query)
    monkeypatch.setattr(search_mod.Search, 'generate_response', fake_generate)

    rv = client.get(f'/{Endpoint.search}?q=whoogle%20!', headers={'Accept': 'application/json'})
    assert rv._status_code == 303
    data = json.loads(rv.data)
    assert data['redirect'] == 'https://example.com/lucky'





"""Google Custom Search Engine (CSE) API Client

This module provides a client for Google's Custom Search JSON API,
allowing users to bring their own API key (BYOK) for search functionality.
"""

import httpx
from typing import Optional
from dataclasses import dataclass
from urllib.parse import urlparse

from flask import render_template


# Google Custom Search API endpoint
CSE_API_URL = 'https://www.googleapis.com/customsearch/v1'


class CSEException(Exception):
    """Exception raised for CSE API errors"""
    def __init__(self, message: str, code: int = 500, is_quota_error: bool = False):
        self.message = message
        self.code = code
        self.is_quota_error = is_quota_error
        super().__init__(self.message)


@dataclass
class CSEError:
    """Represents an error from the CSE API"""
    code: int
    message: str
    
    @property
    def is_quota_exceeded(self) -> bool:
        return self.code == 429 or 'quota' in self.message.lower()
    
    @property
    def is_invalid_key(self) -> bool:
        return self.code == 400 or 'invalid' in self.message.lower()


@dataclass
class CSEResult:
    """Represents a single search result from CSE API"""
    title: str
    link: str
    snippet: str
    display_link: str
    html_title: Optional[str] = None
    html_snippet: Optional[str] = None
    # Image-specific fields (populated for image search)
    image_url: Optional[str] = None
    thumbnail_url: Optional[str] = None
    image_width: Optional[int] = None
    image_height: Optional[int] = None
    context_link: Optional[str] = None  # Page where image was found


@dataclass
class CSEResponse:
    """Represents a complete CSE API response"""
    results: list[CSEResult]
    total_results: str
    search_time: float
    query: str
    start_index: int
    is_image_search: bool = False
    error: Optional[CSEError] = None
    
    @property
    def has_error(self) -> bool:
        return self.error is not None
    
    @property
    def has_results(self) -> bool:
        return len(self.results) > 0


class CSEClient:
    """Client for Google Custom Search Engine API
    
    Usage:
        client = CSEClient(api_key='your-key', cse_id='your-cse-id')
        response = client.search('python programming')
        
        if response.has_error:
            print(f"Error: {response.error.message}")
        else:
            for result in response.results:
                print(f"{result.title}: {result.link}")
    """
    
    def __init__(self, api_key: str, cse_id: str, timeout: float = 10.0):
        """Initialize CSE client
        
        Args:
            api_key: Google API key with Custom Search API enabled
            cse_id: Custom Search Engine ID (cx parameter)
            timeout: Request timeout in seconds
        """
        self.api_key = api_key
        self.cse_id = cse_id
        self.timeout = timeout
        self._client = httpx.Client(timeout=timeout)
    
    def search(
        self,
        query: str,
        start: int = 1,
        num: int = 10,
        safe: str = 'off',
        language: str = '',
        country: str = '',
        search_type: str = ''
    ) -> CSEResponse:
        """Execute a search query against the CSE API
        
        Args:
            query: Search query string
            start: Starting result index (1-based, for pagination)
            num: Number of results to return (max 10)
            safe: Safe search setting ('off', 'medium', 'high')
            language: Language restriction (e.g., 'lang_en')
            country: Country restriction (e.g., 'countryUS')
            search_type: Type of search ('image' for image search, '' for web)
            
        Returns:
            CSEResponse with results or error information
        """
        params = {
            'key': self.api_key,
            'cx': self.cse_id,
            'q': query,
            'start': start,
            'num': min(num, 10),  # API max is 10
            'safe': safe,
        }
        
        # Add search type for image search
        if search_type == 'image':
            params['searchType'] = 'image'
        
        # Add optional parameters
        if language:
            # CSE uses 'lr' for language restrict
            params['lr'] = language
        if country:
            # CSE uses 'cr' for country restrict
            params['cr'] = country
        
        try:
            response = self._client.get(CSE_API_URL, params=params)
            data = response.json()
            
            # Check for API errors
            if 'error' in data:
                error_info = data['error']
                return CSEResponse(
                    results=[],
                    total_results='0',
                    search_time=0.0,
                    query=query,
                    start_index=start,
                    error=CSEError(
                        code=error_info.get('code', 500),
                        message=error_info.get('message', 'Unknown error')
                    )
                )
            
            # Parse successful response
            search_info = data.get('searchInformation', {})
            items = data.get('items', [])
            is_image = search_type == 'image'
            
            results = []
            for item in items:
                # Extract image-specific data if present
                image_data = item.get('image', {})
                
                results.append(CSEResult(
                    title=item.get('title', ''),
                    link=item.get('link', ''),
                    snippet=item.get('snippet', ''),
                    display_link=item.get('displayLink', ''),
                    html_title=item.get('htmlTitle'),
                    html_snippet=item.get('htmlSnippet'),
                    # Image fields
                    image_url=item.get('link') if is_image else None,
                    thumbnail_url=image_data.get('thumbnailLink'),
                    image_width=image_data.get('width'),
                    image_height=image_data.get('height'),
                    context_link=image_data.get('contextLink')
                ))
            
            return CSEResponse(
                results=results,
                total_results=search_info.get('totalResults', '0'),
                search_time=float(search_info.get('searchTime', 0)),
                query=query,
                start_index=start,
                is_image_search=is_image
            )
            
        except httpx.TimeoutException:
            return CSEResponse(
                results=[],
                total_results='0',
                search_time=0.0,
                query=query,
                start_index=start,
                error=CSEError(code=408, message='Request timed out')
            )
        except httpx.RequestError as e:
            return CSEResponse(
                results=[],
                total_results='0',
                search_time=0.0,
                query=query,
                start_index=start,
                error=CSEError(code=500, message=f'Request failed: {str(e)}')
            )
        except Exception as e:
            return CSEResponse(
                results=[],
                total_results='0',
                search_time=0.0,
                query=query,
                start_index=start,
                error=CSEError(code=500, message=f'Unexpected error: {str(e)}')
            )
    
    def close(self):
        """Close the HTTP client"""
        self._client.close()
    
    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self.close()


def cse_results_to_html(response: CSEResponse, query: str) -> str:
    """Convert CSE API response to HTML matching Whoogle's result format
    
    This generates HTML that mimics the structure expected by Whoogle's
    existing filter and result processing pipeline.
    
    Args:
        response: CSEResponse from the API
        query: Original search query
        
    Returns:
        HTML string formatted like Google search results
    """
    if response.has_error:
        error = response.error
        if error.is_quota_exceeded:
            return _error_html(
                'API Quota Exceeded',
                'Your Google Custom Search API quota has been exceeded. '
                'Free tier allows 100 queries/day. Wait until midnight PT '
                'or enable billing in Google Cloud Console.'
            )
        elif error.is_invalid_key:
            return _error_html(
                'Invalid API Key',
                'Your Google Custom Search API key is invalid. '
                'Please check your API key and CSE ID in settings.'
            )
        else:
            return _error_html('Search Error', error.message)
    
    if not response.has_results:
        return _no_results_html(query)
    
    # Use different HTML structure for image vs web results
    if response.is_image_search:
        return _image_results_html(response, query)
    
    # Build HTML results matching Whoogle's expected structure
    results_html = []
    
    for result in response.results:
        # Escape HTML in content
        title = _escape_html(result.title)
        snippet = _escape_html(result.snippet)
        link = result.link
        display_link = _escape_html(result.display_link)
        
        # Use HTML versions if available (they have bold tags for query terms)
        if result.html_title:
            title = result.html_title
        if result.html_snippet:
            snippet = result.html_snippet
        
        # Match the structure used by Google/mock results
        result_html = f'''
        <div class="ZINbbc xpd O9g5cc uUPGi">
            <div class="kCrYT">
                <a href="{link}">
                    <h3 class="BNeawe vvjwJb AP7Wnd">{title}</h3>
                    <div class="BNeawe UPmit AP7Wnd luh4tb" style="color: var(--whoogle-result-url);">{display_link}</div>
                </a>
            </div>
            <div class="kCrYT">
                <div class="BNeawe s3v9rd AP7Wnd">
                    <span class="VwiC3b">{snippet}</span>
                </div>
            </div>
        </div>
        '''
        results_html.append(result_html)
    
    # Build pagination if needed
    pagination_html = ''
    if int(response.total_results) > 10:
        pagination_html = _pagination_html(response.start_index, response.query)
    
    # Wrap in expected structure
    # Add data-cse attribute to prevent collapse_sections from collapsing these results
    return f'''
    <html>
    <body>
        <div id="main" data-cse="true">
            <div id="cnt">
                <div id="rcnt">
                    <div id="center_col">
                        <div id="res">
                            <div id="search">
                                <div id="rso">
                                    {''.join(results_html)}
                                </div>
                            </div>
                        </div>
                        {pagination_html}
                    </div>
                </div>
            </div>
        </div>
    </body>
    </html>
    '''


def _escape_html(text: str) -> str:
    """Escape HTML special characters"""
    if not text:
        return ''
    return (text
            .replace('&', '&amp;')
            .replace('<', '&lt;')
            .replace('>', '&gt;')
            .replace('"', '&quot;')
            .replace("'", '&#39;'))


def _error_html(title: str, message: str) -> str:
    """Generate error HTML"""
    return f'''
    <html>
    <body>
        <div id="main">
            <div style="padding: 20px; text-align: center;">
                <h2 style="color: #d93025;">{_escape_html(title)}</h2>
                <p>{_escape_html(message)}</p>
            </div>
        </div>
    </body>
    </html>
    '''


def _no_results_html(query: str) -> str:
    """Generate no results HTML"""
    return f'''
    <html>
    <body>
        <div id="main">
            <div style="padding: 20px;">
                <p>No results found for <b>{_escape_html(query)}</b></p>
            </div>
        </div>
    </body>
    </html>
    '''


def _image_results_html(response: CSEResponse, query: str) -> str:
    """Generate HTML for image search results using the imageresults template
    
    Args:
        response: CSEResponse with image results
        query: Original search query
        
    Returns:
        HTML string formatted for image results display
    """
    # Convert CSE results to the format expected by imageresults.html template
    results = []
    for result in response.results:
        image_url = result.image_url or result.link
        thumbnail_url = result.thumbnail_url or image_url
        web_page = result.context_link or result.link
        domain = urlparse(web_page).netloc if web_page else result.display_link
        
        results.append({
            'domain': domain,
            'img_url': image_url,
            'web_page': web_page,
            'img_tbn': thumbnail_url
        })
    
    # Build pagination link if needed
    next_link = None
    if int(response.total_results) > response.start_index + len(response.results) - 1:
        next_start = response.start_index + 10
        next_link = f'search?q={query}&tbm=isch&start={next_start}'
    
    # Use the same template as regular image results
    return render_template(
        'imageresults.html',
        length=len(results),
        results=results,
        view_label="View Image",
        next_link=next_link
    )


def _pagination_html(current_start: int, query: str) -> str:
    """Generate pagination links"""
    # CSE API uses 1-based indexing, 10 results per page
    current_page = (current_start - 1) // 10 + 1
    
    prev_link = ''
    next_link = ''
    
    if current_page > 1:
        prev_start = (current_page - 2) * 10 + 1
        prev_link = f'<a href="search?q={query}&start={prev_start}">Previous</a>'
    
    next_start = current_page * 10 + 1
    next_link = f'<a href="search?q={query}&start={next_start}">Next</a>'
    
    return f'''
    <div id="foot" style="text-align: center; padding: 20px;">
        {prev_link}
        <span style="margin: 0 20px;">Page {current_page}</span>
        {next_link}
    </div>
    '''



import pytest

from app import app
from app.request import Request, TorError
from app.models.config import Config


class FakeResponse:
    def __init__(self, text: str = '', status_code: int = 200, content: bytes = b''):
        self.text = text
        self.status_code = status_code
        self.content = content or b''


class FakeHttpClient:
    def __init__(self, tor_ok: bool):
        self._tor_ok = tor_ok

    def get(self, url, headers=None, cookies=None, retries=0, backoff_seconds=0.5, use_cache=False):
        if 'check.torproject.org' in url:
            return FakeResponse(text=('Congratulations' if self._tor_ok else 'Not Tor'))
        return FakeResponse(text='', status_code=200, content=b'OK')

    def close(self):
        pass


def build_config(tor: bool) -> Config:
    # Minimal config with tor flag
    with app.app_context():
        return Config(**{'tor': tor})


def test_tor_validation_success(monkeypatch):
    # Prevent real Tor signal attempts
    monkeypatch.setattr('app.request.send_tor_signal', lambda signal: True)
    cfg = build_config(tor=True)
    req = Request(normal_ua='TestUA', root_path='http://localhost:5000', config=cfg, http_client=FakeHttpClient(tor_ok=True))
    # Avoid sending a Tor NEWNYM/HEARTBEAT in unit tests by setting attempt>0 false path
    resp = req.send(base_url='https://example.com', query='')
    assert req.tor_valid is True
    assert resp.status_code == 200


def test_tor_validation_failure(monkeypatch):
    # Prevent real Tor signal attempts
    monkeypatch.setattr('app.request.send_tor_signal', lambda signal: True)
    cfg = build_config(tor=True)
    req = Request(normal_ua='TestUA', root_path='http://localhost:5000', config=cfg, http_client=FakeHttpClient(tor_ok=False))
    with pytest.raises(TorError):
        _ = req.send(base_url='https://example.com', query='')




import os
import re
from typing import Any
from app.filter import Filter
from app.request import gen_query
from app.utils.misc import get_proxy_host_url
from app.utils.results import get_first_link
from app.services.cse_client import CSEClient, cse_results_to_html
from bs4 import BeautifulSoup as bsoup
from cryptography.fernet import Fernet, InvalidToken
from flask import g

TOR_BANNER = '<hr><h1 style="text-align: center">You are using Tor</h1><hr>'
CAPTCHA = 'div class="g-recaptcha"'


def needs_https(url: str) -> bool:
    """Checks if the current instance needs to be upgraded to HTTPS

    Note that all Heroku instances are available by default over HTTPS, but
    do not automatically set up a redirect when visited over HTTP.

    Args:
        url: The instance url

    Returns:
        bool: True/False representing the need to upgrade

    """
    https_only = bool(os.getenv('HTTPS_ONLY', 0))
    is_heroku = url.endswith('.herokuapp.com')
    is_http = url.startswith('http://')

    return (is_heroku and is_http) or (https_only and is_http)


def has_captcha(results: str) -> bool:
    """Checks to see if the search results are blocked by a captcha

    Args:
        results: The search page html as a string

    Returns:
        bool: True/False indicating if a captcha element was found

    """
    return CAPTCHA in results


class Search:
    """Search query preprocessor - used before submitting the query or
    redirecting to another site

    Attributes:
        request: the incoming flask request
        config: the current user config settings
        session_key: the flask user fernet key
    """
    def __init__(self, request, config, session_key, cookies_disabled=False, user_request=None):
        method = request.method
        self.request = request
        self.request_params = request.args if method == 'GET' else request.form
        self.user_agent = request.headers.get('User-Agent')
        self.feeling_lucky = False
        self.config = config
        self.session_key = session_key
        self.query = ''
        self.widget = ''
        self.cookies_disabled = cookies_disabled
        self.user_request = user_request
        self.search_type = self.request_params.get(
            'tbm') if 'tbm' in self.request_params else ''

    def __getitem__(self, name) -> Any:
        return getattr(self, name)

    def __setitem__(self, name, value) -> None:
        return setattr(self, name, value)

    def __delitem__(self, name) -> None:
        return delattr(self, name)

    def __contains__(self, name) -> bool:
        return hasattr(self, name)

    def new_search_query(self) -> str:
        """Parses a plaintext query into a valid string for submission

        Also decrypts the query string, if encrypted (in the case of
        paginated results).

        Returns:
            str: A valid query string

        """
        q = self.request_params.get('q')

        if q is None or len(q) == 0:
            return ''
        else:
            # Attempt to decrypt if this is an internal link
            try:
                q = Fernet(self.session_key).decrypt(q.encode()).decode()
            except InvalidToken:
                pass

        # Strip '!' for "feeling lucky" queries
        if match := re.search(r"(^|\s)!($|\s)", q):
            self.feeling_lucky = True
            start, end = match.span()
            self.query = " ".join([seg for seg in [q[:start], q[end:]] if seg])
        else:
            self.feeling_lucky = False
            self.query = q

        # Check for possible widgets
        self.widget = "ip" if re.search("([^a-z0-9]|^)my *[^a-z0-9] *(ip|internet protocol)" +
                        "($|( *[^a-z0-9] *(((addres|address|adres|" +
                        "adress)|a)? *$)))", self.query.lower()) else self.widget
        self.widget = 'calculator' if re.search(
                r"\bcalculator\b|\bcalc\b|\bcalclator\b|\bmath\b",
                self.query.lower()) else self.widget
        return self.query

    def generate_response(self) -> str:
        """Generates a response for the user's query

        Returns:
            str: A string response to the search query, in the form of a URL
                 or string representation of HTML content.

        """
        mobile = 'Android' in self.user_agent or 'iPhone' in self.user_agent
        # reconstruct url if X-Forwarded-Host header present
        root_url = get_proxy_host_url(
            self.request,
            self.request.url_root,
            root=True)

        content_filter = Filter(self.session_key,
                                root_url=root_url,
                                mobile=mobile,
                                config=self.config,
                                query=self.query,
                                page_url=self.request.url)
        
        # Check if CSE (Custom Search Engine) should be used
        use_cse = (
            self.config.use_cse and 
            self.config.cse_api_key and 
            self.config.cse_id
        )
        
        if use_cse:
            # Use Google Custom Search API
            return self._generate_cse_response(content_filter, root_url, mobile)
        
        # Default: Use traditional scraping method
        return self._generate_scrape_response(content_filter, root_url, mobile)
    
    def _generate_cse_response(self, content_filter: Filter, root_url: str, mobile: bool) -> str:
        """Generate response using Google Custom Search API
        
        Args:
            content_filter: Filter instance for processing results
            root_url: Root URL of the instance
            mobile: Whether this is a mobile request
            
        Returns:
            str: HTML response string
        """
        # Get pagination start index from request params
        start = int(self.request_params.get('start', 1))
        
        # Determine safe search setting
        safe = 'high' if self.config.safe else 'off'
        
        # Determine search type (web or image)
        # tbm=isch or udm=2 indicates image search
        search_type = ''
        if self.search_type == 'isch' or self.request_params.get('udm') == '2':
            search_type = 'image'
        
        # Create CSE client and perform search
        with CSEClient(
            api_key=self.config.cse_api_key,
            cse_id=self.config.cse_id
        ) as client:
            response = client.search(
                query=self.query,
                start=start,
                safe=safe,
                language=self.config.lang_search,
                country=self.config.country,
                search_type=search_type
            )
        
        # Convert CSE response to HTML
        html_content = cse_results_to_html(response, self.query)
        
        # Store full query for tabs
        self.full_query = self.query
        
        # Parse and filter the HTML
        html_soup = bsoup(html_content, 'html.parser')
        
        # Handle feeling lucky
        if self.feeling_lucky:
            if response.has_results and response.results:
                return response.results[0].link
            self.feeling_lucky = False
        
        # Apply content filter (encrypts links, applies CSS, etc.)
        formatted_results = content_filter.clean(html_soup)
        
        return str(formatted_results)
    
    def _generate_scrape_response(self, content_filter: Filter, root_url: str, mobile: bool) -> str:
        """Generate response using traditional HTML scraping
        
        Args:
            content_filter: Filter instance for processing results
            root_url: Root URL of the instance
            mobile: Whether this is a mobile request
            
        Returns:
            str: HTML response string
        """
        full_query = gen_query(self.query,
                               self.request_params,
                               self.config)
        self.full_query = full_query

        # force mobile search when view image is true and
        # the request is not already made by a mobile
        is_image_query = ('tbm=isch' in full_query) or ('udm=2' in full_query)
        # Always parse image results when hitting the images endpoint (udm=2)
        # to avoid Google returning only text/AI blocks.
        view_image = is_image_query

        client = self.user_request or g.user_request
        get_body = client.send(query=full_query,
                               force_mobile=self.config.view_image,
                               user_agent=self.user_agent)

        # Produce cleanable html soup from response
        get_body_safed = get_body.text.replace("&lt;","andlt;").replace("&gt;","andgt;")
        html_soup = bsoup(get_body_safed, 'html.parser')
        
        # Ensure we extract only the content within <html> if it exists
        # This prevents doctype declarations from appearing in the output
        if html_soup.html:
            html_soup = html_soup.html

        # Replace current soup if view_image is active
        if view_image:
            html_soup = content_filter.view_image(html_soup)

        # Indicate whether or not a Tor connection is active
        if (self.user_request or g.user_request).tor_valid:
            html_soup.insert(0, bsoup(TOR_BANNER, 'html.parser'))

        formatted_results = content_filter.clean(html_soup)
        if self.feeling_lucky:
            if lucky_link := get_first_link(formatted_results):
                return lucky_link

            # Fall through to regular search if unable to find link
            self.feeling_lucky = False

        # Append user config to all search links, if available
        param_str = ''.join('&{}={}'.format(k, v)
                            for k, v in
                            self.request_params.to_dict(flat=True).items()
                            if self.config.is_safe_key(k))
        for link in formatted_results.find_all('a', href=True):
            link['rel'] = "nofollow noopener noreferrer"
            if 'search?' not in link['href'] or link['href'].index(
                    'search?') > 1:
                continue
            link['href'] += param_str

        return str(formatted_results)



"""
User Agent Generator for Opera-based UA strings.

This module generates realistic Opera User Agent strings based on patterns
found in working UA strings that successfully bypass Google's restrictions.
"""

import json
import os
import random
from datetime import datetime, timedelta
from typing import List, Dict


# Default fallback UA if generation fails
DEFAULT_FALLBACK_UA = "Opera/9.80 (iPad; Opera Mini/5.0.17381/503; U; eu) Presto/2.6.35 Version/11.10)"

# Opera UA Pattern Templates
OPERA_PATTERNS = [
    # Opera Mini (J2ME/MIDP)
    "Opera/9.80 (J2ME/MIDP; Opera Mini/{version}/{build}; U; {lang}) Presto/{presto} Version/{final}",
    
    # Opera Mobile (Android)
    "Opera/9.80 (Android; Linux; Opera Mobi/{build}; U; {lang}) Presto/{presto} Version/{final}",
    
    # Opera Mobile (iPhone)
    "Opera/9.80 (iPhone; Opera Mini/{version}/{build}; U; {lang}) Presto/{presto} Version/{final}",
    
    # Opera Mobile (iPad)
    "Opera/9.80 (iPad; Opera Mini/{version}/{build}; U; {lang}) Presto/{presto} Version/{final}",
]

# Randomization pools based on working UAs
OPERA_MINI_VERSIONS = [
    "4.0", "4.1.11321", "4.1.12965", "4.1.13573", "4.1.13907", "4.1.14287", 
    "4.1.15082", "4.2.13057", "4.2.13221", "4.2.13265", "4.2.13337", 
    "4.2.13400", "4.2.13918", "4.2.13943", "4.2.14320", "4.2.14409", 
    "4.2.14753", "4.2.14881", "4.2.14885", "4.2.14912", "4.2.15066",
    "4.2.15410", "4.2.16007", "4.2.16320", "4.2.18887", "4.2.19634",
    "4.2.21465", "4.2.22228", "4.2.23453", "4.2.24721", "4.3.13337",
    "4.3.24214", "4.4.26736", "4.4.29476", "4.5.33867", "4.5.40312",
    "5.0.15650", "5.0.16823", "5.0.17381", "5.0.17443", "5.0.18635",
    "5.0.18741", "5.0.19683", "5.0.19693", "5.0.20873", "5.0.22349",
    "5.1.21051", "5.1.21126", "5.1.21214", "5.1.21415", "5.1.21594",
    "5.1.21595", "5.1.22296", "5.1.22303", "5.1.22396", "5.1.22460",
    "5.1.22783", "5.1.22784", "6.0.24095", "6.0.24212", "6.0.24455",
    "6.1.25375", "6.1.25378", "6.1.25759", "6.24093", "6.24096",
    "6.24209", "6.24288", "6.5.26955", "6.5.29702", "7.0.29952",
    "7.1.32052", "7.1.32444", "7.1.32694", "7.29530", "7.5.33361",
    "7.6.35766", "9.80", "36.2.2254"
]

OPERA_MOBI_BUILDS = [
    "27", "49", "447", "498", "1181", "1209", "3730",
    "ADR-1011151731", "ADR-1012211514", "ADR-1012221546", "ADR-1012272315",
    "SYB-1103211396", "SYB-1104061449", "SYB-1107071606",
    "ADR-1111101157"
]

BUILD_NUMBERS = [
    "18.678", "18.684", "18.738", "18.794", "19.892", "19.916",
    "20.2477", "20.2479", "20.2485", "20.2489", "21.529", "22.387",
    "22.394", "22.401", "22.414", "22.453", "22.478", "23.317",
    "23.333", "23.334", "23.377", "23.390", "24.741", "24.743",
    "24.746", "24.783", "24.838", "24.871", "24.899", "25.657",
    "25.677", "25.729", "25.872", "26.1305", "27.1366", "27.1407",
    "27.1573", "28.2075", "28.2555", "28.2647", "28.2766", "29.3594",
    "30.3316", "31.1350", "35.2883", "35.5706", "37.6584", "119.132",
    "170.51", "170.54", "764", "870", "886", "490", "503"
]

PRESTO_VERSIONS = [
    "2.2.0", "2.4.15", "2.4.154.15", "2.4.18", "2.5.25", "2.5.28",
    "2.6.35", "2.7.60", "2.7.81", "2.8.119", "2.8.149", "2.8.191",
    "2.9.201", "2.12.423"
]

FINAL_VERSIONS = [
    "10.00", "10.1", "10.5", "10.54", "10.5454", "11.00", "11.10",
    "12.02", "12.16", "13.00"
]

LANGUAGES = [
    # English variants
    "en", "en-US", "en-GB", "en-CA", "en-AU", "en-NZ", "en-ZA", "en-IN", "en-SG",
    # Western European
    "de", "de-DE", "de-AT", "de-CH",
    "fr", "fr-FR", "fr-CA", "fr-BE", "fr-CH", "fr-LU",
    "es", "es-ES", "es-MX", "es-AR", "es-CO", "es-CL", "es-PE", "es-VE", "es-LA",
    "it", "it-IT", "it-CH",
    "pt", "pt-PT", "pt-BR",
    "nl", "nl-NL", "nl-BE",
    # Nordic languages
    "da", "da-DK",
    "sv", "sv-SE",
    "no", "no-NO", "nb", "nn",
    "fi", "fi-FI",
    "is", "is-IS",
    # Eastern European
    "pl", "pl-PL",
    "cs", "cs-CZ",
    "sk", "sk-SK",
    "hu", "hu-HU",
    "ro", "ro-RO",
    "bg", "bg-BG",
    "hr", "hr-HR",
    "sr", "sr-RS",
    "sl", "sl-SI",
    "uk", "uk-UA",
    "ru", "ru-RU",
    # Asian languages
    "zh", "zh-CN", "zh-TW", "zh-HK",
    "ja", "ja-JP",
    "ko", "ko-KR",
    "th", "th-TH",
    "vi", "vi-VN",
    "id", "id-ID",
    "ms", "ms-MY",
    "fil", "tl",
    # Middle Eastern
    "tr", "tr-TR",
    "ar", "ar-SA", "ar-AE", "ar-EG",
    "he", "he-IL",
    "fa", "fa-IR",
    # Other
    "hi", "hi-IN",
    "bn", "bn-IN",
    "ta", "ta-IN",
    "te", "te-IN",
    "mr", "mr-IN",
    "el", "el-GR",
    "ca", "ca-ES",
    "eu", "eu-ES"
]



def generate_opera_ua() -> str:
    """
    Generate a single random Opera User Agent string.
    
    Returns:
        str: A randomly generated Opera UA string
    """
    pattern = random.choice(OPERA_PATTERNS)
    
    # Determine which parameters to use based on the pattern
    params = {
        'lang': random.choice(LANGUAGES)
    }
    
    if '{version}' in pattern:
        params['version'] = random.choice(OPERA_MINI_VERSIONS)
    
    if '{build}' in pattern:
        # Use MOBI build for "Opera Mobi", regular build for "Opera Mini"
        if "Opera Mobi" in pattern:
            params['build'] = random.choice(OPERA_MOBI_BUILDS)
        else:
            params['build'] = random.choice(BUILD_NUMBERS)
    
    if '{presto}' in pattern:
        params['presto'] = random.choice(PRESTO_VERSIONS)
    
    if '{final}' in pattern:
        params['final'] = random.choice(FINAL_VERSIONS)
    
    return pattern.format(**params)


def generate_ua_pool(count: int = 10) -> List[str]:
    """
    Generate a pool of unique Opera User Agent strings.
    
    Args:
        count: Number of UA strings to generate (default: 10)
    
    Returns:
        List[str]: List of unique UA strings
    """
    ua_pool = set()
    
    # Keep generating until we have enough unique UAs
    # Add safety limit to prevent infinite loop
    max_attempts = count * 100
    attempts = 0
    
    try:
        while len(ua_pool) < count and attempts < max_attempts:
            ua = generate_opera_ua()
            ua_pool.add(ua)
            attempts += 1
    except Exception:
        # If generation fails entirely, return at least the default fallback
        if not ua_pool:
            return [DEFAULT_FALLBACK_UA]
    
    # If we couldn't generate enough, fill remaining with default
    result = list(ua_pool)
    while len(result) < count:
        result.append(DEFAULT_FALLBACK_UA)
    
    return result


def save_ua_pool(uas: List[str], cache_path: str) -> None:
    """
    Save UA pool to cache file.
    
    Args:
        uas: List of UA strings to save
        cache_path: Path to cache file
    """
    cache_data = {
        'generated_at': datetime.now().isoformat(),
        'user_agents': uas
    }
    
    # Ensure directory exists
    cache_dir = os.path.dirname(cache_path)
    if cache_dir and not os.path.exists(cache_dir):
        os.makedirs(cache_dir, exist_ok=True)
    
    with open(cache_path, 'w', encoding='utf-8') as f:
        json.dump(cache_data, f, indent=2)


def load_custom_ua_list(file_path: str) -> List[str]:
    """
    Load custom UA list from a text file.
    
    Args:
        file_path: Path to text file containing UA strings (one per line)
    
    Returns:
        List[str]: List of UA strings, or empty list if file is invalid
    """
    try:
        with open(file_path, 'r', encoding='utf-8') as f:
            uas = [line.strip() for line in f if line.strip()]
        
        # Validate that we have at least one UA
        if not uas:
            return []
        
        return uas
    except (FileNotFoundError, PermissionError, UnicodeDecodeError):
        return []


def load_ua_pool(cache_path: str, count: int = 10) -> List[str]:
    """
    Load UA pool from custom list file, cache, or generate new one.
    
    Priority order:
    1. Custom UA list file (if WHOOGLE_UA_LIST_FILE is set)
    2. Cached auto-generated UAs
    3. Newly generated UAs
    
    Args:
        cache_path: Path to cache file
        count: Number of UAs to generate if cache is invalid (default: 10)
    
    Returns:
        List[str]: List of UA strings
    """
    # Check for custom UA list file first (highest priority)
    custom_ua_file = os.environ.get('WHOOGLE_UA_LIST_FILE', '').strip()
    if custom_ua_file:
        custom_uas = load_custom_ua_list(custom_ua_file)
        if custom_uas:
            # Custom list loaded successfully
            return custom_uas
        else:
            # Custom file specified but invalid, log warning and fall back
            print(f"Warning: Custom UA list file '{custom_ua_file}' not found or invalid, falling back to auto-generated UAs")
    
    # Check if we should use cache
    use_cache = os.environ.get('WHOOGLE_UA_CACHE_PERSISTENT', '1') == '1'
    refresh_days = int(os.environ.get('WHOOGLE_UA_CACHE_REFRESH_DAYS', '0'))
    
    # If cache disabled, always generate new
    if not use_cache:
        uas = generate_ua_pool(count)
        save_ua_pool(uas, cache_path)
        return uas
    
    # Try to load from cache
    if os.path.exists(cache_path):
        try:
            with open(cache_path, 'r', encoding='utf-8') as f:
                cache_data = json.load(f)
            
            # Check if cache is expired (if refresh_days > 0)
            if refresh_days > 0:
                generated_at = datetime.fromisoformat(cache_data['generated_at'])
                age_days = (datetime.now() - generated_at).days
                
                if age_days >= refresh_days:
                    # Cache expired, generate new
                    uas = generate_ua_pool(count)
                    save_ua_pool(uas, cache_path)
                    return uas
            
            # Cache is valid, return it
            return cache_data['user_agents']
        except (json.JSONDecodeError, KeyError, ValueError):
            # Cache file is corrupted, generate new
            pass
    
    # No valid cache, generate new
    uas = generate_ua_pool(count)
    save_ua_pool(uas, cache_path)
    return uas


def get_random_ua(ua_pool: List[str]) -> str:
    """
    Get a random UA from the pool.
    
    Args:
        ua_pool: List of UA strings
    
    Returns:
        str: Random UA string from the pool
    """
    if not ua_pool:
        # Fallback to generating one if pool is empty
        try:
            return generate_opera_ua()
        except Exception:
            # If generation fails, use default fallback
            return DEFAULT_FALLBACK_UA
    
    return random.choice(ua_pool)




import threading
import time
from typing import Any, Dict, Optional, Tuple

import httpx
from cachetools import TTLCache
import ssl
import os

# Import h2 exceptions for better error handling
try:
    from h2.exceptions import ProtocolError as H2ProtocolError
except ImportError:
    H2ProtocolError = None


class HttpxClient:
    """Thin wrapper around httpx.Client providing simple retries and optional TTL caching.

    The client is intended to be safe for reuse across requests. Per-request
    overrides for headers/cookies are supported.
    """

    def __init__(
            self,
            proxies: Optional[Dict[str, str]] = None,
            timeout_seconds: float = 15.0,
            cache_ttl_seconds: int = 30,
            cache_maxsize: int = 256,
            http2: bool = True) -> None:
        # Allow disabling HTTP/2 via environment variable
        # HTTP/2 can sometimes cause protocol errors with certain servers
        if os.environ.get('WHOOGLE_DISABLE_HTTP2', '').lower() in ('1', 'true', 't', 'yes', 'y'):
            http2 = False
            
        client_kwargs = dict(http2=http2,
                             timeout=timeout_seconds,
                             follow_redirects=True)
        # Prefer future-proof mounts when proxies are provided; fall back to proxies=
        self._proxies = proxies or {}
        self._http2 = http2

        # Determine verify behavior and initialize client with fallbacks
        self._verify = self._determine_verify_setting()
        try:
            self._client = self._build_client(client_kwargs, self._verify)
        except ssl.SSLError:
            # Fallback to system trust store
            try:
                system_ctx = ssl.create_default_context()
                self._client = self._build_client(client_kwargs, system_ctx)
                self._verify = system_ctx
            except ssl.SSLError:
                insecure_fallback = os.environ.get('WHOOGLE_INSECURE_FALLBACK', '0').lower() in ('1', 'true', 't', 'yes', 'y')
                if insecure_fallback:
                    self._client = self._build_client(client_kwargs, False)
                    self._verify = False
                else:
                    raise
        self._timeout_seconds = timeout_seconds
        self._cache = TTLCache(maxsize=cache_maxsize, ttl=cache_ttl_seconds)
        self._cache_lock = threading.Lock()

    def _determine_verify_setting(self):
        """Determine SSL verification setting from environment.

        Honors:
        - WHOOGLE_CA_BUNDLE: path to CA bundle file
        - WHOOGLE_SSL_VERIFY: '0' to disable verification
        - WHOOGLE_SSL_BACKEND: 'system' to prefer system trust store
        """
        ca_bundle = os.environ.get('WHOOGLE_CA_BUNDLE', '').strip()
        if ca_bundle:
            return ca_bundle

        verify_env = os.environ.get('WHOOGLE_SSL_VERIFY', '1').lower()
        if verify_env in ('0', 'false', 'no', 'n'):
            return False

        backend = os.environ.get('WHOOGLE_SSL_BACKEND', '').lower()
        if backend == 'system':
            return ssl.create_default_context()

        return True

    def _build_client(self, client_kwargs: Dict[str, Any], verify: Any) -> httpx.Client:
        """Construct httpx.Client with proxies and provided verify setting."""
        kwargs = dict(client_kwargs)
        kwargs['verify'] = verify
        if self._proxies:
            proxy_values = list(self._proxies.values())
            single_proxy = proxy_values[0] if proxy_values and all(v == proxy_values[0] for v in proxy_values) else None
            if single_proxy:
                try:
                    return httpx.Client(proxy=single_proxy, **kwargs)
                except TypeError:
                    try:
                        return httpx.Client(proxies=self._proxies, **kwargs)
                    except TypeError:
                        mounts: Dict[str, httpx.Proxy] = {}
                        for scheme_key, url in self._proxies.items():
                            prefix = f"{scheme_key}://"
                            mounts[prefix] = httpx.Proxy(url)
                        return httpx.Client(mounts=mounts, **kwargs)
            else:
                try:
                    return httpx.Client(proxies=self._proxies, **kwargs)
                except TypeError:
                    mounts: Dict[str, httpx.Proxy] = {}
                    for scheme_key, url in self._proxies.items():
                        prefix = f"{scheme_key}://"
                        mounts[prefix] = httpx.Proxy(url)
                    return httpx.Client(mounts=mounts, **kwargs)
        else:
            return httpx.Client(**kwargs)

    @property
    def proxies(self) -> Dict[str, str]:
        return self._proxies

    def _cache_key(self, method: str, url: str, headers: Optional[Dict[str, str]]) -> Tuple[str, str, Tuple[Tuple[str, str], ...]]:
        normalized_headers = tuple(sorted((headers or {}).items()))
        return (method.upper(), url, normalized_headers)

    def get(self,
            url: str,
            headers: Optional[Dict[str, str]] = None,
            cookies: Optional[Dict[str, str]] = None,
            retries: int = 2,
            backoff_seconds: float = 0.5,
            use_cache: bool = False) -> httpx.Response:
        if use_cache:
            key = self._cache_key('GET', url, headers)
            with self._cache_lock:
                cached = self._cache.get(key)
            if cached is not None:
                return cached

        last_exc: Optional[Exception] = None
        attempt = 0
        while attempt <= retries:
            try:
                # Check if client is closed and recreate if needed
                if self._client.is_closed:
                    self._recreate_client()
                    
                response = self._client.get(url, headers=headers, cookies=cookies)
                if use_cache and response.status_code == 200:
                    with self._cache_lock:
                        self._cache[key] = response
                return response
            except Exception as exc:
                last_exc = exc
                # Check for specific errors that require client recreation
                should_recreate = False
                
                if isinstance(exc, (httpx.HTTPError, RuntimeError)):
                    if "client has been closed" in str(exc).lower():
                        should_recreate = True
                
                # Handle H2 protocol errors (connection state issues)
                if H2ProtocolError and isinstance(exc, H2ProtocolError):
                    should_recreate = True
                
                # Also check if the error message contains h2 protocol error info
                if "ProtocolError" in str(exc) or "ConnectionState.CLOSED" in str(exc):
                    should_recreate = True
                
                if should_recreate:
                    self._recreate_client()
                    if attempt < retries:
                        time.sleep(backoff_seconds * (2 ** attempt))
                        attempt += 1
                        continue
                
                # For non-recoverable errors or last attempt, raise
                if attempt == retries:
                    raise
                    
                # For other errors, still retry with backoff
                time.sleep(backoff_seconds * (2 ** attempt))
                attempt += 1

        # Should not reach here
        if last_exc:
            raise last_exc
        raise httpx.HTTPError('Unknown HTTP error')

    def _recreate_client(self) -> None:
        """Recreate the HTTP client when it has been closed."""
        try:
            self._client.close()
        except Exception:
            pass  # Client might already be closed
        
        # Recreate with same configuration
        client_kwargs = dict(timeout=self._timeout_seconds,
                             follow_redirects=True,
                             http2=self._http2)

        try:
            self._client = self._build_client(client_kwargs, self._verify)
        except ssl.SSLError:
            try:
                system_ctx = ssl.create_default_context()
                self._client = self._build_client(client_kwargs, system_ctx)
                self._verify = system_ctx
            except ssl.SSLError:
                insecure_fallback = os.environ.get('WHOOGLE_INSECURE_FALLBACK', '0').lower() in ('1', 'true', 't', 'yes', 'y')
                if insecure_fallback:
                    self._client = self._build_client(client_kwargs, False)
                    self._verify = False
                else:
                    raise

    def close(self) -> None:
        self._client.close()





import os
from typing import Dict, Tuple

from app.services.http_client import HttpxClient


_clients: Dict[tuple, HttpxClient] = {}


def _proxies_key(proxies: Dict[str, str]) -> Tuple[Tuple[str, str], Tuple[str, str]]:
    if not proxies:
        return tuple(), tuple()
    # Separate http/https for stable key
    items = sorted((proxies or {}).items())
    return tuple(items), tuple(items)


def get_http_client(proxies: Dict[str, str]) -> HttpxClient:
    # Determine HTTP/2 enablement from env (default on)
    http2_env = os.environ.get('WHOOGLE_HTTP2', '1').lower()
    http2_enabled = http2_env in ('1', 'true', 't', 'yes', 'y')

    key = (_proxies_key(proxies or {}), http2_enabled)
    client = _clients.get(key)
    if client is not None:
        return client
    client = HttpxClient(proxies=proxies or None, http2=http2_enabled)
    _clients[key] = client
    return client


def close_all_clients() -> None:
    for client in list(_clients.values()):
        try:
            client.close()
        except Exception:
            pass
    _clients.clear()





from app import app
from app.request import Request
from app.models.config import Config


class FakeHttpClient:
    def get(self, url, headers=None, cookies=None, retries=0, backoff_seconds=0.5, use_cache=False):
        # Minimal XML in Google Toolbar Autocomplete format
        xml = (
            '<?xml version="1.0"?>\n'
            '<topp>\n'
            '  <CompleteSuggestion><suggestion data="whoogle"/></CompleteSuggestion>\n'
            '  <CompleteSuggestion><suggestion data="whoogle search"/></CompleteSuggestion>\n'
            '</topp>'
        )
        class R:
            text = xml
        return R()

    def close(self):
        pass


def test_autocomplete_parsing():
    with app.app_context():
        cfg = Config(**{})
    req = Request(normal_ua='UA', root_path='http://localhost:5000', config=cfg, http_client=FakeHttpClient())
    suggestions = req.autocomplete('who')
    assert 'whoogle' in suggestions
    assert 'whoogle search' in suggestions




import cssutils
from bs4 import BeautifulSoup
from bs4.element import ResultSet, Tag
from cryptography.fernet import Fernet
from flask import render_template
import html
import urllib.parse as urlparse
import os
from urllib.parse import parse_qs, urlencode, urlunparse
import re

from app.models.g_classes import GClasses
from app.request import VALID_PARAMS, MAPS_URL
from app.utils.misc import get_abs_url, read_config_bool
from app.utils.results import (
    BLANK_B64, GOOG_IMG, GOOG_STATIC, G_M_LOGO_URL, LOGO_URL, SITE_ALTS,
    has_ad_content, filter_link_args, append_anon_view, get_site_alt,
)
from app.models.endpoint import Endpoint
from app.models.config import Config


MAPS_ARGS = ['q', 'daddr']

minimal_mode_sections = ['Top stories', 'Images', 'People also ask']
unsupported_g_pages = [
    'support.google.com',
    'accounts.google.com',
    'policies.google.com',
    'google.com/preferences',
    'google.com/intl',
    'advanced_search',
    'tbm=shop',
    'ageverification.google.co.kr'
]

unsupported_g_divs = ['google.com/preferences?hl=', 'ageverification.google.co.kr']


def extract_q(q_str: str, href: str) -> str:
    """Extracts the 'q' element from a result link. This is typically
    either the link to a result's website, or a string.

    Args:
        q_str: The result link to parse
        href: The full url to check for standalone 'q' elements first,
              rather than parsing the whole query string and then checking.

    Returns:
        str: The 'q' element of the link, or an empty string
    """
    return parse_qs(q_str, keep_blank_values=True)['q'][0] if ('&q=' in href or '?q=' in href) else ''


def build_map_url(href: str) -> str:
    """Tries to extract known args that explain the location in the url. If a
    location is found, returns the default url with it. Otherwise, returns the
    url unchanged.

    Args:
        href: The full url to check.

    Returns:
        str: The parsed url, or the url unchanged.
    """
    # parse the url
    parsed_url = parse_qs(href)
    # iterate through the known parameters and try build the url
    for param in MAPS_ARGS:
        if param in parsed_url:
            return MAPS_URL + "?q=" + parsed_url[param][0]

    # query could not be extracted returning unchanged url
    return href


def clean_query(query: str) -> str:
    """Strips the blocked site list from the query, if one is being
    used.

    Args:
        query: The query string

    Returns:
        str: The query string without any "-site:..." filters
    """
    return query[:query.find('-site:')] if '-site:' in query else query


def clean_css(css: str, page_url: str) -> str:
    """Removes all remote URLs from a CSS string.

    Args:
        css: The CSS string

    Returns:
        str: The filtered CSS, with URLs proxied through Whoogle
    """
    sheet = cssutils.parseString(css)
    urls = cssutils.getUrls(sheet)

    for url in urls:
        abs_url = get_abs_url(url, page_url)
        if abs_url.startswith('data:'):
            continue
        css = css.replace(
            url,
            f'{Endpoint.element}?type=image/png&url={abs_url}'
        )

    return css


class Filter:
    # Minimum number of child div elements that indicates a collapsible section
    # Regular search results typically have fewer child divs (< 7)
    # Special sections like "People also ask", "Related searches" have more (>= 7)
    # This threshold helps identify and collapse these extended result sections
    RESULT_CHILD_LIMIT = 7

    def __init__(
            self,
            user_key: str,
            config: Config,
            root_url='',
            page_url='',
            query='',
            mobile=False) -> None:
        self.soup = None
        self.config = config
        self.mobile = mobile
        self.user_key = user_key
        self.page_url = page_url
        self.query = query
        self.main_divs = ResultSet('')
        self._elements = 0
        self._av = set()

        self.root_url = root_url[:-1] if root_url.endswith('/') else root_url

    def __getitem__(self, name):
        return getattr(self, name)

    @property
    def elements(self):
        return self._elements

    def encrypt_path(self, path, is_element=False) -> str:
        # Encrypts path to avoid plaintext results in logs
        if is_element:
            # Element paths are encrypted separately from text, to allow key
            # regeneration once all items have been served to the user
            enc_path = Fernet(self.user_key).encrypt(path.encode()).decode()
            self._elements += 1
            return enc_path

        return Fernet(self.user_key).encrypt(path.encode()).decode()

    def clean(self, soup) -> BeautifulSoup:
        self.soup = soup
        self.main_divs = self.soup.find('div', {'id': 'main'})
        self.remove_ads()
        self.remove_ai_overview()
        self.remove_block_titles()
        self.remove_block_url()
        self.collapse_sections()
        self.update_css()
        self.update_styling()
        self.remove_block_tabs()

        # self.main_divs is only populated for the main page of search results
        # (i.e. not images/news/etc).
        if self.main_divs:
            for div in self.main_divs:
                self.sanitize_div(div)

        for img in [_ for _ in self.soup.find_all('img') if 'src' in _.attrs]:
            self.update_element_src(img, 'image/png')

        for audio in [_ for _ in self.soup.find_all('audio') if 'src' in _.attrs]:
            self.update_element_src(audio, 'audio/mpeg')
            audio['controls'] = ''

        for link in self.soup.find_all('a', href=True):
            self.update_link(link)
            self.add_favicon(link)

        if self.config.alts:
            self.site_alt_swap()

        input_form = self.soup.find('form')
        if input_form is not None:
            input_form['method'] = 'GET' if self.config.get_only else 'POST'
            # Use a relative URI for submissions
            input_form['action'] = 'search'

        # Ensure no extra scripts passed through
        for script in self.soup('script'):
            script.decompose()

        # Update default footer and header
        footer = self.soup.find('footer')
        if footer:
            # Remove divs that have multiple links beyond just page navigation
            [_.decompose() for _ in footer.find_all('div', recursive=False)
             if len(_.find_all('a', href=True)) > 3]
            for link in footer.find_all('a', href=True):
                link['href'] = f'{link["href"]}&preferences={self.config.preferences}'

        header = self.soup.find('header')
        if header:
            header.decompose()
        # Remove broken "Dark theme" toggle snippets that occasionally slip
        # into the footer.
        self.remove_dark_theme_toggle(self.soup)
        self.remove_site_blocks(self.soup)
        return self.soup

    def sanitize_div(self, div) -> None:
        """Removes escaped script and iframe tags from results

        Returns:
            None (The soup object is modified directly)
        """
        if not div or not isinstance(div, Tag):
            return

        for d in div.find_all('div', recursive=True):
            d_text = d.find(string=True, recursive=False)

            # Ensure we're working with tags that contain text content
            if not d_text or not d.string:
                continue

            d.string = html.unescape(d_text)
            div_soup = BeautifulSoup(d.string, 'html.parser')

            # Remove all valid script or iframe tags in the div
            for script in div_soup.find_all('script'):
                script.decompose()

            for iframe in div_soup.find_all('iframe'):
                iframe.decompose()

            d.string = str(div_soup)

    def add_favicon(self, link) -> None:
        """Adds icons for each returned result, using the result site's favicon

        Returns:
            None (The soup object is modified directly)
        """
        # Skip empty, parentless, or internal links
        show_favicons = read_config_bool('WHOOGLE_SHOW_FAVICONS', True)
        is_valid_link = link and link.parent and link['href'].startswith('http')
        if not show_favicons or not is_valid_link:
            return

        parent = link.parent
        is_result_div = False

        # Check each parent to make sure that the div doesn't already have a
        # favicon attached, and that the div is a result div
        while parent:
            p_cls = parent.attrs.get('class') or []
            if 'has-favicon' in p_cls or GClasses.scroller_class in p_cls:
                return
            elif GClasses.result_class_a not in p_cls:
                parent = parent.parent
            else:
                is_result_div = True
                break

        if not is_result_div:
            return

        # Construct the html for inserting the icon into the parent div
        parsed = urlparse.urlparse(link['href'])
        favicon = self.encrypt_path(
            f'{parsed.scheme}://{parsed.netloc}/favicon.ico',
            is_element=True)
        src = f'{self.root_url}/{Endpoint.element}?url={favicon}' + \
            '&type=image/x-icon'
        html = f'<img class="site-favicon" src="{src}">'

        favicon = BeautifulSoup(html, 'html.parser')
        link.parent.insert(0, favicon)

        # Update all parents to indicate that a favicon has been attached
        parent = link.parent
        while parent:
            p_cls = parent.get('class') or []
            p_cls.append('has-favicon')
            parent['class'] = p_cls
            parent = parent.parent

            if GClasses.result_class_a in p_cls:
                break

    def remove_dark_theme_toggle(self, soup: BeautifulSoup) -> None:
        """Removes stray Dark theme toggle/link fragments that can appear
        in the footer."""
        for node in soup.find_all(string=re.compile(r'Dark theme', re.I)):
            try:
                parent = node.find_parent(
                    lambda tag: tag.name in ['div', 'span', 'p', 'a', 'li',
                                             'section'])
                target = parent or node.parent
                if target:
                    target.decompose()
                else:
                    node.extract()
            except Exception:
                continue

    def remove_site_blocks(self, soup) -> None:
        if not self.config.block or not soup.body:
            return
        search_string = ' '.join(['-site:' +
                                 _ for _ in self.config.block.split(',')])
        selected = soup.body.find_all(string=re.compile(search_string))

        for result in selected:
            result.string.replace_with(result.string.replace(
                                       search_string, ''))

    def remove_ai_overview(self) -> None:
        """Removes Google's AI Overview/SGE results from search results

        Returns:
            None (The soup object is modified directly)
        """
        if not self.main_divs:
            return

        # Patterns that identify AI Overview sections
        ai_patterns = [
            'AI Overview',
            'AI responses may include mistakes',
        ]

        # Result div classes - check both original Google classes and mapped ones
        # since this runs before CSS class replacement
        result_classes = [GClasses.result_class_a]  # 'ZINbbc'
        result_classes.extend(GClasses.result_classes.get(
            GClasses.result_class_a, []))  # ['Gx5Zad']

        # Collect divs to remove first to avoid modifying while iterating
        divs_to_remove = []

        for div in self.main_divs.find_all('div', recursive=True):
            # Check if this div or its children contain AI Overview markers
            div_text = div.get_text()
            if any(pattern in div_text for pattern in ai_patterns):
                # Walk up to find the top-level result div
                parent = div
                while parent:
                    p_cls = parent.attrs.get('class') or []
                    if any(rc in p_cls for rc in result_classes):
                        if parent not in divs_to_remove:
                            divs_to_remove.append(parent)
                        break
                    parent = parent.parent

        # Remove collected divs
        for div in divs_to_remove:
            div.decompose()

    def remove_ads(self) -> None:
        """Removes ads found in the list of search result divs

        Returns:
            None (The soup object is modified directly)
        """
        if not self.main_divs:
            return

        for div in [_ for _ in self.main_divs.find_all('div', recursive=True)]:
            div_ads = [_ for _ in div.find_all('span', recursive=True)
                       if has_ad_content(_.text)]
            _ = div.decompose() if len(div_ads) else None

    def remove_block_titles(self) -> None:
        if not self.main_divs or not self.config.block_title:
            return
        block_title = re.compile(self.config.block_title)
        for div in [_ for _ in self.main_divs.find_all('div', recursive=True)]:
            block_divs = [_ for _ in div.find_all('h3', recursive=True)
                          if block_title.search(_.text) is not None]
            _ = div.decompose() if len(block_divs) else None

    def remove_block_url(self) -> None:
        if not self.main_divs or not self.config.block_url:
            return
        block_url = re.compile(self.config.block_url)
        for div in [_ for _ in self.main_divs.find_all('div', recursive=True)]:
            block_divs = [_ for _ in div.find_all('a', recursive=True)
                          if block_url.search(_.attrs['href']) is not None]
            _ = div.decompose() if len(block_divs) else None

    def remove_block_tabs(self) -> None:
        if self.main_divs:
            for div in self.main_divs.find_all(
                'div',
                attrs={'class': f'{GClasses.main_tbm_tab}'}
            ):
                _ = div.decompose()
        else:
            # when in images tab
            for div in self.soup.find_all(
                'div',
                attrs={'class': f'{GClasses.images_tbm_tab}'}
            ):
                _ = div.decompose()

    def collapse_sections(self) -> None:
        """Collapses long result sections ("people also asked", "related
         searches", etc) into "details" elements

        These sections are typically the only sections in the results page that
        have more than ~5 child divs within a primary result div.

        Returns:
            None (The soup object is modified directly)
        """
        minimal_mode = read_config_bool('WHOOGLE_MINIMAL')

        def pull_child_divs(result_div: BeautifulSoup):
            try:
                top_level_divs = result_div.find_all('div', recursive=False)
                if not top_level_divs:
                    return []
                return top_level_divs[0].find_all('div', recursive=False)
            except Exception:
                return []

        if not self.main_divs:
            return

        # Skip collapsing for CSE (Custom Search Engine) results
        # CSE results have a data-cse attribute on the main container
        if self.soup.find(attrs={'data-cse': 'true'}):
            return

        # Loop through results and check for the number of child divs in each
        for result in self.main_divs.find_all():
            result_children = pull_child_divs(result)
            if minimal_mode:
                if any(f">{x}</span" in str(s) for s in result_children
                   for x in minimal_mode_sections):
                    result.decompose()
                    continue
                for s in result_children:
                    if ('Twitter ›' in str(s)):
                        result.decompose()
                        continue
                if len(result_children) < self.RESULT_CHILD_LIMIT:
                    continue
            else:
                if len(result_children) < self.RESULT_CHILD_LIMIT:
                    continue

            # Find and decompose the first element with an inner HTML text val.
            # This typically extracts the title of the section (i.e. "Related
            # Searches", "People also ask", etc)
            # If there are more than one child tags with text
            # parenthesize the rest except the first
            label = 'Collapsed Results'
            subtitle = None
            for elem in result_children:
                if elem.text:
                    content = list(elem.strings)
                    label = content[0]
                    if len(content) > 1:
                        subtitle = '<span> (' + \
                            ''.join(content[1:]) + ')</span>'
                    elem.decompose()
                    break

            # Create the new details element to wrap around the result's
            # first parent
            parent = None
            idx = 0
            while not parent and idx < len(result_children):
                parent = result_children[idx].parent
                idx += 1

            details = BeautifulSoup(features='html.parser').new_tag('details')
            summary = BeautifulSoup(features='html.parser').new_tag('summary')
            summary.string = label

            if subtitle:
                soup = BeautifulSoup(subtitle, 'html.parser')
                summary.append(soup)

            details.append(summary)

            if parent and not minimal_mode:
                parent.wrap(details)
            elif parent and minimal_mode:
                # Remove parent element from document if "minimal mode" is
                # enabled
                parent.decompose()

    def update_element_src(self, element: Tag, mime: str, attr='src') -> None:
        """Encrypts the original src of an element and rewrites the element src
        to use the "/element?src=" pass-through.

        Returns:
            None (The soup element is modified directly)

        """
        src = element[attr].split(' ')[0]

        if src.startswith('//'):
            src = 'https:' + src
        elif src.startswith('data:'):
            return

        if src.startswith(LOGO_URL):
            # Re-brand with Whoogle logo
            element.replace_with(BeautifulSoup(
                render_template('logo.html'),
                features='html.parser'))
            return
        elif src.startswith(G_M_LOGO_URL):
            # Re-brand with single-letter Whoogle logo
            element['src'] = 'static/img/favicon/apple-icon.png'
            element.parent['href'] = 'home'
            return
        elif src.startswith(GOOG_IMG) or GOOG_STATIC in src:
            element['src'] = BLANK_B64
            return

        element[attr] = f'{self.root_url}/{Endpoint.element}?url=' + (
            self.encrypt_path(
                src,
                is_element=True
            ) + '&type=' + urlparse.quote(mime)
        )

    def update_css(self) -> None:
        """Updates URLs used in inline styles to be proxied by Whoogle
        using the /element endpoint.

        Returns:
            None (The soup element is modified directly)

        """
        # Filter all <style> tags
        for style in self.soup.find_all('style'):
            style.string = clean_css(style.string, self.page_url)

        # TODO: Convert remote stylesheets to style tags and proxy all
        # remote requests
        # for link in soup.find_all('link', attrs={'rel': 'stylesheet'}):
            # print(link)

    def update_styling(self) -> None:
        # Update CSS classes for result divs
        soup = GClasses.replace_css_classes(self.soup)

        # Remove unnecessary button(s)
        for button in self.soup.find_all('button'):
            button.decompose()

        # Remove svg logos
        for svg in self.soup.find_all('svg'):
            svg.decompose()

        # Update logo
        logo = self.soup.find('a', {'class': 'l'})
        if logo and self.mobile:
            logo['style'] = ('display:flex; justify-content:center; '
                             'align-items:center; color:#685e79; '
                             'font-size:18px; ')

        # Fix search bar length on mobile
        try:
            search_bar = self.soup.find('header').find('form').find('div')
            search_bar['style'] = 'width: 100%;'
        except AttributeError:
            pass

        # Fix body max width on images tab
        style = self.soup.find('style')
        div = self.soup.find('div', attrs={
            'class': f'{GClasses.images_tbm_tab}'})
        if style and div and not self.mobile:
            css = style.string
            css_html_tag = (
                'html{'
                'font-family: Roboto, Helvetica Neue, Arial, sans-serif;'
                'font-size: 14px;'
                'line-height: 20px;'
                'text-size-adjust: 100%;'
                'word-wrap: break-word;'
                '}'
            )
            css = f"{css_html_tag}{css}"
            css = re.sub('body{(.*?)}',
                         'body{padding:0 12px;margin:0 auto;max-width:1200px;}',
                         css)
            style.string = css

        # Normalize the max width between result types so the page doesn't
        # jump in size when switching tabs.
        if not self.mobile:
            max_width_css = (
                'body, #cnt, #center_col, .main, .e9EfHf, #searchform, '
                '.GyAeWb, .s6JM6d {'
                'max-width:1200px;'
                'margin:0 auto;'
                'padding-left:12px;'
                'padding-right:12px;'
                '}'
            )
            # Build the style tag using a fresh soup to avoid cases where the
            # current soup lacks the helper methods (e.g., non-root elements).
            factory_soup = BeautifulSoup('', 'html.parser')
            extra_style = factory_soup.new_tag('style')
            extra_style.string = max_width_css
            if self.soup.head:
                self.soup.head.append(extra_style)
            else:
                self.soup.insert(0, extra_style)

    def update_link(self, link: Tag) -> None:
        """Update internal link paths with encrypted path, otherwise remove
        unnecessary redirects and/or marketing params from the url

        Args:
            link: A bs4 Tag element to inspect and update

        Returns:
            None (the tag is updated directly)

        """
        parsed_link = urlparse.urlparse(link['href'])
        if '/url?q=' in link['href']:
            link_netloc = extract_q(parsed_link.query, link['href'])
        else:
            link_netloc = parsed_link.netloc

        # Remove any elements that direct to unsupported Google pages
        if any(url in link_netloc for url in unsupported_g_pages):
            # Replaces the /url google unsupported link to the direct url
            link['href'] = link_netloc
            parent = link.parent

            if any(divlink in link_netloc for divlink in unsupported_g_divs):
                # Handle case where a search is performed in a different
                # language than what is configured. This usually returns a
                # div with the same classes as normal search results, but with
                # a link to configure language preferences through Google.
                # Since we want all language config done through Whoogle, we
                # can safely decompose this element.
                while parent:
                    p_cls = parent.attrs.get('class') or []
                    if f'{GClasses.result_class_a}' in p_cls:
                        parent.decompose()
                        break
                    parent = parent.parent
            else:
                # Remove cases where google links appear in the footer
                while parent:
                    p_cls = parent.attrs.get('class') or []
                    if parent.name == 'footer' or f'{GClasses.footer}' in p_cls:
                        link.decompose()
                    parent = parent.parent

            if link.decomposed:
                return

        # Replace href with only the intended destination (no "utm" type tags)
        href = link['href'].replace('https://www.google.com', '')
        result_link = urlparse.urlparse(href)
        q = extract_q(result_link.query, href)

        if q.startswith('/') and q not in self.query and 'spell=1' not in href:
            # Internal google links (i.e. mail, maps, etc) should still
            # be forwarded to Google
            link['href'] = 'https://google.com' + q
        elif q.startswith('https://accounts.google.com'):
            # Remove Sign-in link
            link.decompose()
            return
        elif '/search?q=' in href:
            # "li:1" implies the query should be interpreted verbatim,
            # which is accomplished by wrapping the query in double quotes
            if 'li:1' in href:
                q = '"' + q + '"'
            new_search = 'search?q=' + self.encrypt_path(q)

            query_params = parse_qs(urlparse.urlparse(href).query)
            for param in VALID_PARAMS:
                if param not in query_params:
                    continue
                param_val = query_params[param][0]
                new_search += '&' + param + '=' + param_val
            link['href'] = new_search
        elif 'url?q=' in href:
            # Strip unneeded arguments
            link['href'] = filter_link_args(q)

            # Add alternate viewing options for results,
            # if the result doesn't already have an AV link
            netloc = urlparse.urlparse(link['href']).netloc
            if self.config.anon_view and netloc not in self._av:
                self._av.add(netloc)
                append_anon_view(link, self.config)

        else:
            if href.startswith(MAPS_URL):
                # Maps links don't work if a site filter is applied
                link['href'] = build_map_url(link['href'])
            elif (href.startswith('/?') or href.startswith('/search?') or
                  href.startswith('/imgres?')):
                # make sure that tags can be clicked as relative URLs
                link['href'] = href[1:]
            elif href.startswith('/intl/'):
                # do nothing, keep original URL for ToS
                pass
            elif href.startswith('/preferences'):
                # there is no config specific URL, remove this
                link.decompose()
                return
            else:
                link['href'] = href

        if self.config.new_tab and (
            link["href"].startswith("http")
            or link["href"].startswith("imgres?")
        ):
            link["target"] = "_blank"

    def site_alt_swap(self) -> None:
        """Replaces link locations and page elements if "alts" config
        is enabled
        """
        # Precompute regex for sites (escape dots) and common prefixes
        site_keys = list(SITE_ALTS.keys())
        if not site_keys:
            return
        sites_pattern = re.compile('|'.join([re.escape(k) for k in site_keys]))
        prefix_pattern = re.compile(r'^(?:https?:\/\/)?(?:(?:www|mobile|m)\.)?')

        # 1) Replace bare domain divs (single token) once, avoiding duplicates
        for div in self.soup.find_all('div', string=sites_pattern):
            if not div or not div.string:
                continue
            if len(div.string.split(' ')) != 1:
                continue
            match = sites_pattern.search(div.string)
            if not match:
                continue
            site = match.group(0)
            alt = SITE_ALTS.get(site, '')
            if not alt:
                continue
            # Skip if already contains the alt to avoid old.old.* repetition
            if alt in div.string:
                continue
            div.string = div.string.replace(site, alt)

        # 2) Update link hrefs and descriptions in a single pass
        for link in self.soup.find_all('a', href=True):
            link['href'] = get_site_alt(link['href'])

            # Find a description text node matching a known site
            desc_nodes = link.find_all(string=sites_pattern)
            if not desc_nodes:
                continue
            desc_node = desc_nodes[0]
            link_str = str(desc_node)

            # Determine which site key is present in the description
            site_match = sites_pattern.search(link_str)
            if not site_match:
                continue
            site = site_match.group(0)
            alt = SITE_ALTS.get(site, '')
            if not alt:
                continue

            # Avoid duplication if alt already present
            if alt in link_str:
                continue

            # Medium-specific handling remains to avoid matching substrings
            if 'medium.com' in link_str:
                if link_str.startswith('medium.com') or '.medium.com' in link_str:
                    replaced = SITE_ALTS['medium.com'] + link_str[
                        link_str.find('medium.com') + len('medium.com'):
                    ]
                else:
                    replaced = link_str
            else:
                # If the description looks like a URL with scheme, replace only the host
                if '://' in link_str:
                    scheme, rest = link_str.split('://', 1)
                    host, sep, path = rest.partition('/')
                    # Drop common prefixes from host when swapping to a fully-qualified alt
                    alt_parsed = urlparse.urlparse(alt)
                    alt_host = alt_parsed.netloc if alt_parsed.netloc else alt.replace('https://', '').replace('http://', '')
                    # If alt includes a scheme, prefer its host; otherwise use alt as host
                    if alt_parsed.scheme:
                        new_host = alt_host
                    else:
                        # When alt has no scheme, still replace entire host
                        new_host = alt
                    # Prevent replacing if host already equals target
                    if host == new_host:
                        replaced = link_str
                    else:
                        replaced = f"{scheme}://{new_host}{sep}{path}"
                else:
                    # No scheme in the text; include optional prefixes in replacement
                    # Replace any leading www./m./mobile. + site with alt host (no scheme)
                    alt_parsed = urlparse.urlparse(alt)
                    alt_host = alt_parsed.netloc if alt_parsed.netloc else alt.replace('https://', '').replace('http://', '')
                    # Build a pattern that includes optional prefixes for the specific site
                    site_with_prefix = re.compile(rf'(?:(?:www|mobile|m)\.)?{re.escape(site)}')
                    replaced = site_with_prefix.sub(alt_host, link_str, count=1)

            new_desc = BeautifulSoup(features='html.parser').new_tag('div')
            new_desc.string = replaced
            desc_node.replace_with(new_desc)

    def view_image(self, soup) -> BeautifulSoup:
        """Parses image results from Google Images and rewrites them into the
        lightweight Whoogle image results template.

        Google now serves image results via the modern udm=2 endpoint, where
        the raw HTML contains only placeholder thumbnails. The actual image
        URLs live inside serialized data blobs in script tags. We extract that
        data and pair it with the visible result cards.
        """

        def _decode_url(url: str) -> str:
            if not url:
                return ''
            # Decode common escaped characters found in the script blobs
            return html.unescape(
                url.replace('\\u003d', '=').replace('\\u0026', '&')
            )

        def _extract_image_data(modern_soup: BeautifulSoup) -> dict:
            """Extracts docid -> {img_url, img_tbn} from serialized scripts."""
            scripts_text = ' '.join(
                script.string for script in modern_soup.find_all('script')
                if script.string
            )
            pattern = re.compile(
                r'\[0,"(?P<docid>[^"]+)",\["(?P<thumb>https://encrypted-tbn[^"]+)"'
                r'(?:,\d+,\d+)?\],\["(?P<full>https?://[^"]+?)"'
                r'(?:,\d+,\d+)?\]',
                re.DOTALL
            )
            results_map = {}
            for match in pattern.finditer(scripts_text):
                docid = match.group('docid')
                thumb = _decode_url(match.group('thumb'))
                full = _decode_url(match.group('full'))
                results_map[docid] = {
                    'img_tbn': thumb,
                    'img_url': full
                }
            return results_map

        def _parse_modern_results(modern_soup: BeautifulSoup) -> list:
            cards = modern_soup.find_all(
                'div',
                attrs={
                    'data-attrid': 'images universal',
                    'data-docid': True
                }
            )
            if not cards:
                return []

            meta_map = _extract_image_data(modern_soup)
            parsed = []
            seen = set()

            for card in cards:
                docid = card.get('data-docid')
                meta = meta_map.get(docid, {})
                img_url = meta.get('img_url')
                img_tbn = meta.get('img_tbn')

                # Fall back to the inline src if we failed to map the docid
                if not img_tbn:
                    img_tag = card.find('img')
                    if img_tag:
                        candidate_src = img_tag.get('src')
                        if candidate_src and candidate_src.startswith('http'):
                            img_tbn = candidate_src

                web_page = card.get('data-lpage') or ''
                if not web_page:
                    link = card.find('a', href=True)
                    if link:
                        web_page = link['href']

                key = (img_url, img_tbn, web_page)
                if not any(key) or key in seen:
                    continue
                seen.add(key)

                parsed.append({
                    'domain': urlparse.urlparse(web_page).netloc
                    if web_page else '',
                    'img_url': img_url or img_tbn or '',
                    'web_page': web_page,
                    'img_tbn': img_tbn or img_url or ''
                })
            return parsed

        # Try parsing the modern (udm=2) layout first
        modern_results = _parse_modern_results(soup)
        if modern_results:
            # TODO: Implement proper image pagination. Google images uses
            # infinite scroll with `ijn` offsets; we need a clean,
            # de-duplicated pagination strategy before exposing a Next link.
            next_link = None
            return BeautifulSoup(
                render_template(
                    'imageresults.html',
                    length=len(modern_results),
                    results=modern_results,
                    view_label="View Image",
                    next_link=next_link
                ),
                features='html.parser'
            )

        # get some tags that are unchanged between mobile and pc versions
        cor_suggested = soup.find_all('table', attrs={'class': "By0U9"})
        next_pages = soup.find('table', attrs={'class': "uZgmoc"})

        results = []
        # find results div
        results_div = soup.find('div', attrs={'class': "nQvrDb"})
        # find all the results (if any)
        results_all = []
        if results_div:
            results_all = results_div.find_all('div', attrs={'class': "lIMUZd"})

        for item in results_all:
            link = item.find('a', href=True)
            if not link:
                continue

            urls = link['href'].split('&imgrefurl=')

            # Skip urls that are not two-element lists
            if len(urls) != 2:
                continue

            img_url = urlparse.unquote(urls[0].replace(
                f'/{Endpoint.imgres}?imgurl=', ''))

            try:
                # Try to strip out only the necessary part of the web page link
                web_page = urlparse.unquote(urls[1].split('&')[0])
            except IndexError:
                web_page = urlparse.unquote(urls[1])

            img_tag = link.find('img')
            if not img_tag:
                continue

            img_tbn = urlparse.unquote(
                img_tag.get('src') or img_tag.get('data-src', '')
            )

            if not img_tbn:
                continue

            results.append({
                'domain': urlparse.urlparse(web_page).netloc,
                'img_url': img_url,
                'web_page': web_page,
                'img_tbn': img_tbn
            })

        soup = BeautifulSoup(render_template('imageresults.html',
                                             length=len(results),
                                             results=results,
                                             view_label="View Image"),
                             features='html.parser')

        # replace correction suggested by google object if exists
        if len(cor_suggested):
            suggested_tables = soup.find_all(
                'table',
                attrs={'class': "By0U9"}
            )
            if suggested_tables:
                suggested_tables[0].replaceWith(cor_suggested[0])

        # replace next page object at the bottom of the page, when present
        next_page_tables = soup.find_all('table', attrs={'class': "uZgmoc"})
        if next_pages and next_page_tables:
            next_page_tables[0].replaceWith(next_pages)

        # TODO: Reintroduce pagination for legacy image layout if needed.

        return soup



import json

import pytest

from app.models.endpoint import Endpoint
from app.utils import search as search_mod


def test_captcha_json_block(client, monkeypatch):
    def fake_new_query(self):
        self.query = 'test'
        return self.query

    def fake_generate(self):
        # Inject a captcha marker into HTML so route returns 503 JSON
        return '<div>div class="g-recaptcha"</div>'

    monkeypatch.setattr(search_mod.Search, 'new_search_query', fake_new_query)
    monkeypatch.setattr(search_mod.Search, 'generate_response', fake_generate)

    rv = client.get(f'/{Endpoint.search}?q=test&format=json')
    assert rv._status_code == 503
    data = json.loads(rv.data)
    assert data['blocked'] is True
    assert 'error_message' in data




import copy
import os

from bs4 import BeautifulSoup

from app import app
from app.filter import Filter
from app.models.config import Config
from app.utils.session import generate_key
from app.utils import results as results_mod


def build_soup(html: str):
    return BeautifulSoup(html, 'html.parser')


def make_filter(soup: BeautifulSoup):
    secret_key = generate_key()
    with app.app_context():
        cfg = Config(**{'alts': True})
    f = Filter(user_key=secret_key, config=cfg)
    f.soup = soup
    return f


def test_no_duplicate_alt_prefix_reddit(monkeypatch):
    original_site_alts = copy.deepcopy(results_mod.SITE_ALTS)
    try:
        # Simulate user setting alt to old.reddit.com
        monkeypatch.setitem(results_mod.SITE_ALTS, 'reddit.com', 'old.reddit.com')

        html = '''
        <div id="main">
          <a href="https://www.reddit.com/r/whoogle">www.reddit.com</a>
          <div>www.reddit.com</div>
          <div>old.reddit.com</div>
        </div>
        '''
        soup = build_soup(html)
        f = make_filter(soup)
        f.site_alt_swap()

        # Href replaced once
        a = soup.find('a')
        assert a['href'].startswith('https://old.reddit.com')

        # Bare domain replaced, but already-alt text stays unchanged (no old.old...)
        divs = [d.get_text() for d in soup.find_all('div') if d.get_text().strip()]
        assert 'old.reddit.com' in divs
        assert 'old.old.reddit.com' not in ''.join(divs)
    finally:
        results_mod.SITE_ALTS.clear()
        results_mod.SITE_ALTS.update(original_site_alts)


def test_wikipedia_simple_no_lang_param(monkeypatch):
    original_site_alts = copy.deepcopy(results_mod.SITE_ALTS)
    try:
        monkeypatch.setitem(results_mod.SITE_ALTS, 'wikipedia.org', 'https://wikiless.example')

        html = '''
        <div id="main">
          <a href="https://simple.wikipedia.org/wiki/Whoogle">https://simple.wikipedia.org/wiki/Whoogle</a>
          <div>simple.wikipedia.org</div>
        </div>
        '''
        soup = build_soup(html)
        f = make_filter(soup)
        f.site_alt_swap()

        a = soup.find('a')
        # Should be rewritten to the alt host, without ?lang
        assert a['href'].startswith('https://wikiless.example')
        assert '?lang=' not in a['href']

        # Description host replaced once
        text = soup.find('div').get_text()
        assert 'wikiless.example' in text
        assert 'simple.wikipedia.org' not in text
    finally:
        results_mod.SITE_ALTS.clear()
        results_mod.SITE_ALTS.update(original_site_alts)


def test_single_pass_description_replacement(monkeypatch):
    original_site_alts = copy.deepcopy(results_mod.SITE_ALTS)
    try:
        monkeypatch.setitem(results_mod.SITE_ALTS, 'twitter.com', 'https://nitter.example')

        html = '''
        <div id="main">
          <a href="https://twitter.com/whoogle">https://twitter.com/whoogle</a>
          <div>https://www.twitter.com</div>
        </div>
        '''
        soup = build_soup(html)
        f = make_filter(soup)
        f.site_alt_swap()

        a = soup.find('a')
        assert a['href'].startswith('https://nitter.example')

        # Ensure description got host swapped once, no double scheme or duplication
        main_div = soup.find('div', id='main')
        # The description div is the first inner div under #main in this fixture
        text = main_div.find_all('div')[0].get_text().strip()
        assert text.startswith('https://nitter.example')
        assert 'https://https://' not in text
        assert 'nitter.examplenitter.example' not in text
    finally:
        results_mod.SITE_ALTS.clear()
        results_mod.SITE_ALTS.update(original_site_alts)





from app.models.config import Config
from app.utils.misc import read_config_bool
from app.services.provider import get_http_client
from app.utils.ua_generator import load_ua_pool, get_random_ua, DEFAULT_FALLBACK_UA
from defusedxml import ElementTree as ET
import httpx
import urllib.parse as urlparse
import os
from stem import Signal, SocketError
from stem.connection import AuthenticationFailure
from stem.control import Controller
from stem.connection import authenticate_cookie, authenticate_password

MAPS_URL = 'https://maps.google.com/maps'
AUTOCOMPLETE_URL = ('https://suggestqueries.google.com/'
                    'complete/search?client=toolbar&')

# Valid query params
VALID_PARAMS = ['tbs', 'tbm', 'start', 'near', 'source', 'nfpr']


class TorError(Exception):
    """Exception raised for errors in Tor requests.

    Attributes:
        message: a message describing the error that occurred
        disable: optionally disables Tor in the user config (note:
            this should only happen if the connection has been dropped
            altogether).
    """

    def __init__(self, message, disable=False) -> None:
        self.message = message
        self.disable = disable
        super().__init__(message)


def send_tor_signal(signal: Signal) -> bool:
    use_pass = read_config_bool('WHOOGLE_TOR_USE_PASS')

    confloc = './misc/tor/control.conf'
    # Check that the custom location of conf is real.
    temp = os.getenv('WHOOGLE_TOR_CONF', '')
    if os.path.isfile(temp):
        confloc = temp

    # Attempt to authenticate and send signal.
    try:
        with Controller.from_port(port=9051) as c:
            if use_pass:
                with open(confloc, "r") as conf:
                    # Scan for the last line of the file.
                    for line in conf:
                        pass
                    secret = line.strip('\n')
                authenticate_password(c, password=secret)
            else:
                cookie_path = '/var/lib/tor/control_auth_cookie'
                authenticate_cookie(c, cookie_path=cookie_path)
            c.signal(signal)
            os.environ['TOR_AVAILABLE'] = '1'
            return True
    except (SocketError, AuthenticationFailure,
            ConnectionRefusedError, ConnectionError):
        # TODO: Handle Tor authentication (password and cookie)
        os.environ['TOR_AVAILABLE'] = '0'

    return False


def gen_user_agent(config, is_mobile) -> str:
    # If using custom user agent, return the custom string
    if config.user_agent == 'custom' and config.custom_user_agent:
        return config.custom_user_agent

    # If using environment configuration
    if config.user_agent == 'env_conf':
        if is_mobile:
            env_ua = os.getenv('WHOOGLE_USER_AGENT_MOBILE', '')
            if env_ua:
                return env_ua
        else:
            env_ua = os.getenv('WHOOGLE_USER_AGENT', '')
            if env_ua:
                return env_ua
        # If env vars are not set, fall back to Opera UA
        return DEFAULT_FALLBACK_UA

    # If using default user agent - use auto-generated Opera UA pool
    if config.user_agent == 'default':
        try:
            # Try to load UA pool from cache (lazy loading if not in app.config)
            # First check if we have access to Flask app context
            try:
                from flask import current_app
                if hasattr(current_app, 'config') and 'UA_POOL' in current_app.config:
                    ua_pool = current_app.config['UA_POOL']
                else:
                    # Fall back to loading from disk
                    raise ImportError("UA_POOL not in app config")
            except (ImportError, RuntimeError):
                # No Flask context available or UA_POOL not in config, load from disk
                config_path = os.environ.get('CONFIG_VOLUME', 
                                            os.path.join(os.path.dirname(os.path.abspath(__file__)), 
                                                        'static', 'config'))
                cache_path = os.path.join(config_path, 'ua_cache.json')
                ua_pool = load_ua_pool(cache_path, count=10)
            
            return get_random_ua(ua_pool)
        except Exception as e:
            # If anything goes wrong, fall back to default Opera UA
            print(f"Warning: Could not load UA pool, using fallback Opera UA: {e}")
            return DEFAULT_FALLBACK_UA

    # Fallback for backwards compatibility (old configs or invalid user_agent values)
    return DEFAULT_FALLBACK_UA


def gen_query(query, args, config) -> str:
    param_dict = {key: '' for key in VALID_PARAMS}

    # Use :past(hour/day/week/month/year) if available
    # example search "new restaurants :past month"
    lang = ''
    if ':past' in query and 'tbs' not in args:
        time_range = str.strip(query.split(':past', 1)[-1])
        param_dict['tbs'] = '&tbs=' + ('qdr:' + str.lower(time_range[0]))
    elif 'tbs' in args or 'tbs' in config:
        result_tbs = args.get('tbs') if 'tbs' in args else config['tbs']
        param_dict['tbs'] = '&tbs=' + result_tbs

        # Occasionally the 'tbs' param provided by google also contains a
        # field for 'lr', but formatted strangely. This is a rough solution
        # for this.
        #
        # Example:
        # &tbs=qdr:h,lr:lang_1pl
        # -- the lr param needs to be extracted and remove the leading '1'
        result_params = [_ for _ in result_tbs.split(',') if 'lr:' in _]
        if len(result_params) > 0:
            result_param = result_params[0]
            lang = result_param[result_param.find('lr:') + 3:len(result_param)]

    # Ensure search query is parsable
    query = urlparse.quote(query)

    # Pass along type of results (news, images, books, etc)
    if 'tbm' in args:
        param_dict['tbm'] = '&tbm=' + args.get('tbm')
        # Google Images now expects the modern udm=2 layout; force it when
        # requesting images to avoid redirects to the new AI/text layout.
        if args.get('tbm') == 'isch' and 'udm' not in args:
            param_dict['udm'] = '&udm=2'

    # Get results page start value (10 per page, ie page 2 start val = 20)
    if 'start' in args:
        param_dict['start'] = '&start=' + args.get('start')

    # Search for results near a particular city, if available
    if config.near:
        param_dict['near'] = '&near=' + urlparse.quote(config.near)

    # Set language for results (lr) if source isn't set, otherwise use the
    # result language param provided in the results
    if 'source' in args:
        param_dict['source'] = '&source=' + args.get('source')
        param_dict['lr'] = ('&lr=' + ''.join(
            [_ for _ in lang if not _.isdigit()]
        )) if lang else ''
    else:
        param_dict['lr'] = (
            '&lr=' + config.lang_search
        ) if config.lang_search else ''

    # 'nfpr' defines the exclusion of results from an auto-corrected query
    if 'nfpr' in args:
        param_dict['nfpr'] = '&nfpr=' + args.get('nfpr')

    # 'chips' is used in image tabs to pass the optional 'filter' to add to the
    # given search term
    if 'chips' in args:
        param_dict['chips'] = '&chips=' + args.get('chips')

    param_dict['gl'] = (
        '&gl=' + config.country
    ) if config.country else ''
    param_dict['hl'] = (
        '&hl=' + config.lang_interface.replace('lang_', '')
    ) if config.lang_interface else ''
    param_dict['safe'] = '&safe=' + ('active' if config.safe else 'off')

    # Block all sites specified in the user config
    unquoted_query = urlparse.unquote(query)
    for blocked_site in config.block.replace(' ', '').split(','):
        if not blocked_site:
            continue
        block = (' -site:' + blocked_site)
        query += block if block not in unquoted_query else ''

    for val in param_dict.values():
        if not val:
            continue
        query += val

    return query


class Request:
    """Class used for handling all outbound requests, including search queries,
    search suggestions, and loading of external content (images, audio, etc).

    Attributes:
        normal_ua: the user's current user agent
        root_path: the root path of the whoogle instance
        config: the user's current whoogle configuration
    """

    def __init__(self, normal_ua, root_path, config: Config, http_client=None):
        self.search_url = 'https://www.google.com/search?gbv=1&q='
        # Google Images rejects the lightweight gbv=1 interface. Use the
        # modern udm=2 entrypoint specifically for image searches to avoid the
        # "update your browser" interstitial.
        self.image_search_url = 'https://www.google.com/search?udm=2&q='
        # Optionally send heartbeat to Tor to determine availability
        # Only when Tor is enabled in config to avoid unnecessary socket usage
        if config.tor:
            send_tor_signal(Signal.HEARTBEAT)

        self.language = config.lang_search if config.lang_search else ''
        self.country = config.country if config.country else ''

        # For setting Accept-language Header
        self.lang_interface = ''
        if config.accept_language:
            self.lang_interface = config.lang_interface

        self.mobile = bool(normal_ua) and ('Android' in normal_ua
                                           or 'iPhone' in normal_ua)

        # Generate user agent based on config
        self.modified_user_agent = gen_user_agent(config, self.mobile)
        if not self.mobile:
            self.modified_user_agent_mobile = gen_user_agent(config, True)

        # Dedicated modern UA to use when Google rejects legacy ones (e.g. Images)
        self.image_user_agent = (
            'Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
            'AppleWebKit/537.36 (KHTML, like Gecko) '
            'Chrome/127.0.0.0 Safari/537.36'
        )

        # Set up proxy configuration
        proxy_path = os.environ.get('WHOOGLE_PROXY_LOC', '')
        if proxy_path:
            proxy_type = os.environ.get('WHOOGLE_PROXY_TYPE', '')
            proxy_user = os.environ.get('WHOOGLE_PROXY_USER', '')
            proxy_pass = os.environ.get('WHOOGLE_PROXY_PASS', '')
            auth_str = ''
            if proxy_user:
                auth_str = f'{proxy_user}:{proxy_pass}@'

            proxy_str = f'{proxy_type}://{auth_str}{proxy_path}'
            self.proxies = {
                'https': proxy_str,
                'http': proxy_str
            }
        else:
            self.proxies = {
                'http': 'socks5://127.0.0.1:9050',
                'https': 'socks5://127.0.0.1:9050'
            } if config.tor else {}

        self.tor = config.tor
        self.tor_valid = False
        self.root_path = root_path
        # Initialize HTTP client (shared per proxies)
        self.http_client = http_client or get_http_client(self.proxies)

    def __getitem__(self, name):
        return getattr(self, name)

    def autocomplete(self, query) -> list:
        """Sends a query to Google's search suggestion service

        Args:
            query: The in-progress query to send

        Returns:
            list: The list of matches for possible search suggestions

        """
        # Check if autocomplete is disabled via environment variable
        if os.environ.get('WHOOGLE_AUTOCOMPLETE', '1') == '0':
            return []
            
        try:
            ac_query = dict(q=query)
            if self.language:
                ac_query['lr'] = self.language
            if self.country:
                ac_query['gl'] = self.country
            if self.lang_interface:
                ac_query['hl'] = self.lang_interface

            response = self.send(base_url=AUTOCOMPLETE_URL,
                                 query=urlparse.urlencode(ac_query)).text

            if not response:
                return []

            try:
                root = ET.fromstring(response)
                return [_.attrib['data'] for _ in
                        root.findall('.//suggestion/[@data]')]
            except ET.ParseError:
                # Malformed XML response
                return []
        except Exception as e:
            # Log the error but don't crash - autocomplete is non-essential
            print(f"Autocomplete error: {str(e)}")
            return []

    def send(self, base_url='', query='', attempt=0,
             force_mobile=False, user_agent=''):
        """Sends an outbound request to a URL. Optionally sends the request
        using Tor, if enabled by the user.

        Args:
            base_url: The URL to use in the request
            query: The optional query string for the request
            attempt: The number of attempts made for the request
                (used for cycling through Tor identities, if enabled)
            force_mobile: Optional flag to enable a mobile user agent
                (used for fetching full size images in search results)

        Returns:
            Response: The Response object returned by the requests call

        """
        use_client_user_agent = int(os.environ.get('WHOOGLE_USE_CLIENT_USER_AGENT', '0'))
        if user_agent and use_client_user_agent == 1:
            modified_user_agent = user_agent
        else:
            if force_mobile and not self.mobile:
                modified_user_agent = self.modified_user_agent_mobile
            else:
                modified_user_agent = self.modified_user_agent

        # Some Google endpoints (notably Images) now refuse legacy user agents.
        # If an image search is detected and the generated UA isn't Chromium-
        # like, retry with a modern Chrome string to avoid the "update your
        # browser" interstitial.
        if (('tbm=isch' in query) or ('udm=2' in query)) and 'Chrome' not in modified_user_agent:
            modified_user_agent = self.image_user_agent

        headers = {
            'User-Agent': modified_user_agent,
            'Accept': ('text/html,application/xhtml+xml,application/xml;'
                       'q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8'),
            'Accept-Language': 'en-US,en;q=0.9',
            'Accept-Encoding': 'gzip, deflate, br',
            'Connection': 'keep-alive',
            'Cache-Control': 'max-age=0',
            'Pragma': 'no-cache',
            'Upgrade-Insecure-Requests': '1',
            'Sec-Fetch-Site': 'none',
            'Sec-Fetch-Mode': 'navigate',
            'Sec-Fetch-User': '?1',
            'Sec-Fetch-Dest': 'document'
        }
        # Only attach client hints when using a Chromium-like user agent to
        # avoid sending conflicting information that can trigger unsupported
        # browser pages.
        if 'Chrome' in headers['User-Agent']:
            headers.update({
                'Sec-CH-UA': (
                    '"Not/A)Brand";v="8", '
                    '"Chromium";v="127", '
                    '"Google Chrome";v="127"'
                ),
                'Sec-CH-UA-Mobile': '?0',
                'Sec-CH-UA-Platform': '"Windows"'
            })

 
        # Add Accept-Language header tied to the current config if requested
        if self.lang_interface:
            headers['Accept-Language'] = (
                self.lang_interface.replace('lang_', '') + ';q=1.0'
            )

        # Consent cookies keep Google from showing the interstitial consent wall
        consent_cookies = {
            'CONSENT': 'PENDING+987',
            'SOCS': 'CAESHAgBEhIaAB'
        }

        # Validate Tor conn and request new identity if the last one failed
        if self.tor and not send_tor_signal(
                Signal.NEWNYM if attempt > 0 else Signal.HEARTBEAT):
            raise TorError(
                "Tor was previously enabled, but the connection has been "
                "dropped. Please check your Tor configuration and try again.",
                disable=True)

        # Make sure that the tor connection is valid, if enabled
        if self.tor:
            try:
                tor_check = self.http_client.get('https://check.torproject.org/',
                                                 headers=headers,
                                                 retries=1)
                self.tor_valid = 'Congratulations' in tor_check.text

                if not self.tor_valid:
                    raise TorError(
                        "Tor connection succeeded, but the connection could "
                        "not be validated by torproject.org",
                        disable=True)
            except httpx.RequestError:
                raise TorError(
                    "Error raised during Tor connection validation",
                    disable=True)

        search_base = base_url or self.search_url
        if not base_url and ('tbm=isch' in query or 'udm=2' in query):
            search_base = self.image_search_url

        try:
            response = self.http_client.get(
                search_base + query,
                headers=headers,
                cookies=consent_cookies)
        except httpx.HTTPError as e:
            raise

        # Retry query with new identity if using Tor (max 10 attempts)
        if 'form id="captcha-form"' in response.text and self.tor:
            attempt += 1
            if attempt > 10:
                raise TorError("Tor query failed -- max attempts exceeded 10")
            return self.send(search_base, query, attempt)

        return response



import types

import httpx
import pytest

from app.services.http_client import HttpxClient


def test_httpxclient_follow_redirects_and_proxy(monkeypatch):
    calls = []

    class FakeClient:
        def __init__(self, *args, **kwargs):
            calls.append(kwargs)
        def get(self, *args, **kwargs):
            class R:
                status_code = 200
                text = ''
            return R()
        def close(self):
            pass

    monkeypatch.setattr(httpx, 'Client', FakeClient)

    proxies = {'http': 'socks5://127.0.0.1:9050', 'https': 'socks5://127.0.0.1:9050'}
    client = HttpxClient(proxies=proxies)

    # Ensure the constructor attempted to set follow_redirects and one of proxy/proxies
    assert len(calls) == 1
    kwargs = calls[0]
    assert kwargs.get('follow_redirects') is True
    assert ('proxy' in kwargs) or ('proxies' in kwargs) or ('mounts' in kwargs)

